# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

from pyincore import BaseAnalysis
import pandas as pd
import numpy as np
import time
from typing import List
from pandas import DataFrame
from pyomo.environ import ConcreteModel, Set, Var, Param, Objective, Constraint
from pyomo.environ import quicksum, minimize, maximize, NonNegativeReals
from pyomo.environ import sum_product
import pyomo.environ as pyo
from pyomo.opt import SolverStatus, TerminationCondition
from pyomo.util.infeasible import log_infeasible_constraints


class MultiObjectiveRetrofitOptimization(BaseAnalysis):
    """
        This analysis computes a series of linear programming models for single- and multi-objective
        optimization related to the effect of extreme weather on a community in terms of three objective functions.
        The three objectives used in this program are to minimize economic loss, minimize population dislocation,
        and maximize building functionality. The computation proceeds by iteratively solving constrained linear
        models using epsilon steps.

        The output of the computation a collection of optimal resource allocations.

        Contributors
            | Science: Charles Nicholson, Yunjie Wen
            | Implementation: Dale Cochran , Tarun Adluri , Jorge Duarte, Santiago Núñez-Corrales, Diego Calderon
                              and NCSA IN-CORE Dev Team

        Related publications

        Args:

            incore_client (IncoreClient): Service authentication.

        """

    # Column descriptors
    __Q_col = 'Q_t_hat'
    __Q_rs_col = 'Q_t_hat_rs'
    __SC_col = 'Sc'
    __SC_rs_col = 'Sc_rs'

    __budget_default = 0.2

    def __init__(self, incore_client):
        super(MultiObjectiveRetrofitOptimization, self).__init__(incore_client)

    def run(self):
        """Execute the multiobjective retrofit optimization analysis using parameters and input data."""
        # Read parameters
        model_solver = self.get_parameter('model_solver')
        num_epsilon_steps = self.get_parameter('num_epsilon_steps')

        budget_available = self.__budget_default
        if self.get_parameter('max_budget') != 'default':
            budget_available = self.get_parameter('budget_available')

        inactive_submodels = []

        in_subm = self.get_parameter('inactive_submodels')

        if in_subm is not None:
            inactive_submodels = in_subm

        # Perform code scaling
        scaling_factor = 1.0
        if self.get_parameter('scale_data'):
            scaling_factor = self.get_parameter('scaling_factor')

        building_related_data = self.get_input_dataset('building_related_data').get_dataframe_from_csv()
        strategy_costs = self.get_input_dataset('strategy_costs_data').get_dataframe_from_csv()

        # Convert Z columns to text in both datasets
        building_related_data['Z'] = building_related_data['Z'].astype(str)
        strategy_costs['Z'] = strategy_costs['Z'].astype(str)

        self.multiobjective_retrofit_optimization_model(model_solver, num_epsilon_steps, budget_available,
                                                        scaling_factor, inactive_submodels, building_related_data,
                                                        strategy_costs)

    def multiobjective_retrofit_optimization_model(self, model_solver, num_epsilon_steps, budget_available,
                                                   scaling_factor, inactive_submodels, building_related_data,
                                                   strategy_costs):
        """Performs the computation of the model.

        Args:

            model_solver (str): model solver to use for analysis
            num_epsilon_steps (int): number of epsilon values for the multistep optimization algorithm
            budget_available (float): budget constraint of the optimization analysis
            scaling_factor (float): scaling factor for Q and Sc matrices
            inactive_submodels (list): submodels to avoid during the computation
            building_related_data (pd.DataFrame): building repairs after a disaster event
            strategy_costs (pd.DataFrame): strategy cost data per building

        """
        # Setup the stream that will collect data
        self.ostr = open('pyincore.txt', 'w')

        # Setup the model
        model, sum_sc = self.configure_model(budget_available, scaling_factor, building_related_data,
                                             strategy_costs)
        self.configure_model_objectives(model)
        print("With constraints model")
        self.configure_model_retrofit_costs(model) # Suspicious

        # Choose the solver setting
        if model_solver == "gurobi" or model_solver is None:
            model_solver_setting = pyo.SolverFactory('gurobi', solver_io="python")
        else:
            model_solver_setting = pyo.SolverFactory(model_solver)

        # Solve each model individually
        print("With individual model")
        obj_list = self.solve_individual_models(model, model_solver_setting, sum_sc)
        print("Epsilon values")
        self.configure_min_max_epsilon_values(model, obj_list, num_epsilon_steps)

        print("Epsilon model")
        xresults_df, yresults_df = self.solve_epsilon_models(model, model_solver_setting, inactive_submodels)

        df_list = self.compute_optimal_results(inactive_submodels, xresults_df, yresults_df)

        self.set_result_csv_data("optimal_solution_dv_x", df_list[0], name="optimal_solution_dv_x",
                                 source="dataframe")
        self.set_result_csv_data("optimal_solution_dv_y", df_list[1], name="optimal_solution_dv_y",
                                 source="dataframe")
        return True

    def configure_model(self, budget_available, scaling_factor, building_related_data, strategy_costs):
        """ Configure the base model to perform the multiobjective optimization.

        Args:

            budget_available (float): available budget
            scaling_factor (float): value to scale monetary input data
            building_related_data (DataFrame): table containing building functionality data
            strategy_costs (DataFrame): table containing retrofit strategy costs data

        Returns:
            ConcreteModel: a base, parameterized cost/functionality model

        """
        # Rescale data
        if scaling_factor != 1.0:
            building_related_data[self.__Q_col] = \
                building_related_data[self.__Q_col].map(lambda a: a / scaling_factor)
            strategy_costs[self.__SC_col] = strategy_costs[self.__SC_col].map(lambda a: a / scaling_factor)

        # Setup pyomo
        model = ConcreteModel()

        model.Z = Set(initialize=building_related_data.Z.unique())
        model.S = Set(initialize=building_related_data.S.unique())
        model.K = Set(initialize=building_related_data.K.unique())
        model.K_prime = Set(initialize=strategy_costs["K'"].unique())

        zsk = []
        for y in range(len(building_related_data)):
            i = building_related_data.loc[y, 'Z']  # Identify the i ∈ Z value.
            j = building_related_data.loc[y, 'S']  # Identify the j ∈ S value.
            k = building_related_data.loc[y, 'K']  # Identify the k ∈ K value.
            zsk.append((i, j, k))  # Add the combination to the list.
        zsk = sorted(set(zsk), key=zsk.index)  # Convert the list to an ordered set for Pyomo.
        model.ZSK = Set(initialize=zsk)  # Define and initialize the ZSK set in Pyomo.

        zs = []
        for y in range(len(building_related_data)):
            i = building_related_data.loc[y, 'Z']  # Identify the i ∈ Z value.
            j = building_related_data.loc[y, 'S']  # Identify the j ∈ S value.
            zs.append((i, j))  # Add the combination to the list.
        zs = sorted(set(zs), key=zs.index)  # Convert the list to an ordered set for Pyomo.
        model.ZS = Set(initialize=zs)  # Define and initialize the ZS set in Pyomo.

        kk_prime = []
        for y in range(len(strategy_costs)):
            k = strategy_costs.loc[y, 'K']  # Identify the k ∈ K value.
            k_prime = strategy_costs.loc[y, "K'"]  # Identify the k ∈ K value.
            kk_prime.append((k, k_prime))  # Add the combination to the list.
        kk_prime = sorted(set(kk_prime), key=kk_prime.index)  # Convert the list to an ordered set for Pyomo.
        model.KK_prime = Set(initialize=kk_prime)  # Define and initialize the KK_prime set in Pyomo.

        k_primek = []
        for y in range(len(strategy_costs)):
            k = strategy_costs.loc[y, 'K']  # Identify the k ∈ K value.
            k_prime = strategy_costs.loc[y, "K'"]  # Identify the k ∈ K value.
            if k_prime <= k:
                k_primek.append((k_prime, k))  # Add the combination to the list.
        k_primek = sorted(set(k_primek), key=k_primek.index)  # Convert the list to an ordered set for Pyomo.
        model.K_primeK = Set(initialize=k_primek)  # Define and initialize the K_primeK set in Pyomo.

        # Define the set of all ZSKK' combinations:
        zskk_prime = []
        for y in range(len(strategy_costs)):
            i = strategy_costs.loc[y, 'Z']  # Identify the i ∈ Z value.
            j = strategy_costs.loc[y, 'S']  # Identify the j ∈ S value.
            k = strategy_costs.loc[y, 'K']  # Identify the k ∈ K value.
            k_prime = strategy_costs.loc[y, "K'"]  # Identify the k ∈ K value.
            zskk_prime.append((i, j, k, k_prime))  # Add the combination to the list.
        zskk_prime = sorted(set(zskk_prime), key=zskk_prime.index)  # Convert the list to an ordered set for Pyomo.
        model.ZSKK_prime = Set(initialize=zskk_prime)  # Define and initialize the ZSKK_prime set in Pyomo.
        model.zskk_prime = zskk_prime # Use the redundant index for later reference

        ####################################################################################################
        # DEFINE VARIABLES AND PARAMETERS:
        ####################################################################################################
        # Declare the decision variable x_ijk (total # buildings in zone i of structure type
        # j at code level k after retrofitting):
        model.x_ijk = Var(model.ZSK, within=NonNegativeReals)

        # Declare the decision variable y_ijkk_prime
        # (total # buildings in zone i of structure type j retrofitted from code level k to code level k_prime):
        model.y_ijkk_prime = Var(model.ZSKK_prime, within=NonNegativeReals)

        # Declare economic loss cost parameter l_ijk:
        model.l_ijk = Param(model.ZSK, within=NonNegativeReals, mutable=True)
        for y in range(len(building_related_data)):
            i = building_related_data.loc[y, 'Z']
            j = building_related_data.loc[y, 'S']
            k = building_related_data.loc[y, 'K']
            model.l_ijk[i, j, k] = building_related_data.loc[y, 'l']

        # Declare dislocation parameter d_ijk:
        model.d_ijk = Param(model.ZSK, within=NonNegativeReals, mutable=True)
        for y in range(len(building_related_data)):
            i = building_related_data.loc[y, 'Z']
            j = building_related_data.loc[y, 'S']
            k = building_related_data.loc[y, 'K']
            model.d_ijk[i, j, k] = building_related_data.loc[y, 'd_ijk']

        # Declare the number of buildings parameter b_ijk:
        model.b_ijk = Param(model.ZSK, within=NonNegativeReals, mutable=True)
        for y in range(len(building_related_data)):
            i = building_related_data.loc[y, 'Z']
            j = building_related_data.loc[y, 'S']
            k = building_related_data.loc[y, 'K']
            model.b_ijk[i, j, k] = building_related_data.loc[y, 'b']

        # Declare the building functionality parameter Q_t_hat:
        model.Q_t_hat = Param(model.ZSK, within=NonNegativeReals, mutable=True)
        for y in range(len(building_related_data)):
            i = building_related_data.loc[y, 'Z']
            j = building_related_data.loc[y, 'S']
            k = building_related_data.loc[y, 'K']
            model.Q_t_hat[i, j, k] = building_related_data.loc[y, 'Q_t_hat']

        # Declare the retrofit cost parameter Sc_ijkk':
        model.Sc_ijkk_prime = Param(model.ZSKK_prime, within=NonNegativeReals, mutable=True)
        for y in range(len(strategy_costs)):
            i = strategy_costs.loc[y, 'Z']
            j = strategy_costs.loc[y, 'S']
            k = strategy_costs.loc[y, 'K']
            k_prime = strategy_costs.loc[y, "K'"]
            model.Sc_ijkk_prime[i, j, k, k_prime] = strategy_costs.loc[y, 'Sc']

        ####################################################################################################
        # DECLARE THE TOTAL MAX BUDGET AND TOTAL AVAILABLE BUDGET:
        ####################################################################################################
        # Define the total max budget based on user's input:
        model.B = Param(mutable=True, within=NonNegativeReals)

        if budget_available == self.__budget_default:
            sumSc = quicksum(
                pyo.value(model.Sc_ijkk_prime[i, j, k, 3]) * pyo.value(model.b_ijk[i, j, k]) for i, j, k in model.ZSK)
        else:
            sumSc = budget_available
        # Define the total available budget based on user's input:
        model.B = sumSc * budget_available

        return model, sumSc

    def configure_model_objectives(self, model):
        """ Configure the model by adding objectives

        Args:
            model (ConcreteModel): a base cost/functionality model

        Returns:
            ConcreteModel: a model extended with objective functions

        """
        model.objective_1 = Objective(rule=self.obj_economic, sense=minimize)
        model.econ_loss = Param(mutable=True, within=NonNegativeReals)  # ,default=10000000000)

        model.objective_2 = Objective(rule=self.obj_dislocation, sense=minimize)
        model.dislocation = Param(mutable=True, within=NonNegativeReals)  # ,default=30000)

        model.objective_3 = Objective(rule=self.obj_functionality, sense=maximize)
        model.functionality = Param(mutable=True, within=NonNegativeReals)  # ,default=1)

    def configure_model_retrofit_costs(self, model):
        model.retrofit_budget_constraint = Constraint(rule=self.retrofit_cost_rule)
        model.number_buildings_ij_constraint = Constraint(model.ZS, rule=self.number_buildings_ij_rule)
        model.a = Param(mutable=True)
        model.c = Param(mutable=True)
        model.building_level_constraint = Constraint(model.ZSK, rule=self.building_level_rule)

    def solve_individual_models(self, model, model_solver_setting, sum_sc):
        print("Max Budget: $", sum_sc)
        print("Available Budget: $", pyo.value(model.B))
        print("")

        rlist_obj_1 = self.solve_model_1(model, model_solver_setting)
        rlist_obj_2 = self.solve_model_2(model, model_solver_setting)
        rlist_obj_3 = self.solve_model_3(model, model_solver_setting)

        values_list = [sum_sc, pyo.value(model.B)] + rlist_obj_1 + rlist_obj_2 + rlist_obj_3

        no_epsilon_constr_init_results = pd.DataFrame(data={
            "Label": ["Max Budget: $", "Budget (20% of max)", "Economic loss min epsilon (optimal value)",
                      "Dislocation when optimizing Economic Loss", "Functionality when optimizing Economic Loss",
                      "Dislocation min epsilon (optimal value)", "Economic Loss when optimizing Dislocation",
                      "Functionality when optimizing Dislocation", "Functionality max epsilon (optimal value)",
                      "Economic Loss when optimizing Functionality", "Dislocation when optimizing Functionality"]
            , "Value": values_list})
        filename = 'no_epsilon_constr_init_results_' + str(time.strftime("%m-%d-%Y")) + '.csv'
        no_epsilon_constr_init_results.to_csv(filename)

        return rlist_obj_1 + rlist_obj_2 + rlist_obj_3

    def solve_model_1(self, model, model_solver_setting):
        starttime = time.time()
        print("Initial solve for objective function 1 starting.")
        # Activate objective function 1 (minimize economic loss) and deactivate others:
        model.objective_1.activate()
        model.objective_2.deactivate()
        model.objective_3.deactivate()

        # Solve the model:
        results = model_solver_setting.solve(model)

        # Save the results if the solver returns an optimal solution:
        if (results.solver.status == SolverStatus.ok) and (
                results.solver.termination_condition == TerminationCondition.optimal):
            self.extract_optimization_results(model)
            obj_1_min_epsilon = pyo.value(model.objective_1)  # Save the optimal economic loss value.
            obj_2_value_1 = pyo.value(model.dislocation)  # Save the dislocation value when optimizing economic loss.
            obj_3_value_1 = pyo.value(
                model.functionality)  # Save the functionality value when optimizing economic loss.
            print("Initial solve for objective function 1 complete.")
            print("Economic Loss: ", pyo.value(model.econ_loss),
                  "Dislocation: ", pyo.value(model.dislocation),
                  "Functionality: ", pyo.value(model.functionality))

            print("Economic loss min epsilon (optimal value):", obj_1_min_epsilon)
            print("Dislocation when optimizing Economic Loss:", obj_2_value_1)
            print("Functionality when optimizing Economic Loss:", obj_3_value_1)
            print("")
            results_list = [obj_1_min_epsilon, obj_2_value_1, obj_3_value_1]
        else:
            print("Not Optimal")
            results_list = [0, 0, 0]

        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time for initial obj 1 solve: ", elapsedtime)
        print("")

        return results_list

    def solve_model_2(self, model, model_solver_setting):
        starttime = time.time()
        print("Initial solve for objective function 2 starting.")
        # Activate objective function 2 (minimize dislocation) and deactivate others:
        model.objective_1.deactivate()
        model.objective_2.activate()
        model.objective_3.deactivate()

        # Solve the model:
        results = model_solver_setting.solve(model)

        # Save the results if the solver returns an optimal solution:
        if (results.solver.status == SolverStatus.ok) and (
                results.solver.termination_condition == TerminationCondition.optimal):
            self.extract_optimization_results(model)
            obj_2_min_epsilon = pyo.value(model.objective_2)  # Save the optimal dislocation value.
            obj_1_value_2 = pyo.value(model.econ_loss)  # Save the economic loss value when optimizing dislocation.
            obj_3_value_2 = pyo.value(model.functionality)  # Save the functionality value when optimizing dislocation.
            print("Initial solve for objective function 2 complete.")
            print("Economic Loss: ", pyo.value(model.econ_loss),
                  "Dislocation: ", pyo.value(model.dislocation),
                  "Functionality: ", pyo.value(model.functionality))

            print("Dislocation min epsilon (optimal value):", obj_2_min_epsilon)
            print("Economic Loss when optimizing Dislocation:", obj_1_value_2)
            print("Functionality when optimizing Dislocation:", obj_3_value_2)
            print("")
            results_list = [obj_2_min_epsilon, obj_1_value_2, obj_3_value_2]
        else:
            print("Not Optimal")
            results_list = [0, 0, 0]

        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time for initial obj 2 solve: ", elapsedtime)
        print("")

        return results_list

    def solve_model_3(self, model, model_solver_setting):
        starttime = time.time()
        print("Initial solve for objective function 3 starting.")
        # Activate objective function 3 (maximize functionality) and deactivate others:
        model.objective_1.deactivate()
        model.objective_2.deactivate()
        model.objective_3.activate()

        # Solve the model:
        results = model_solver_setting.solve(model)

        # Save the results if the solver returns an optimal solution:
        if (results.solver.status == SolverStatus.ok) and (
                results.solver.termination_condition == TerminationCondition.optimal):
            self.extract_optimization_results(model)
            obj_3_max_epsilon = pyo.value(model.objective_3)  # Save the optimal functionality value.
            obj_1_value_3 = pyo.value(model.econ_loss)  # Save the economic loss value when optimizing functionality.
            obj_2_value_3 = pyo.value(model.dislocation)  # Save the dislocation value when optimizing functionality.
            print("Initial solve for objective function 3 complete.")
            print("Economic Loss: ", pyo.value(model.econ_loss),
                  "Dislocation: ", pyo.value(model.dislocation),
                  "Functionality: ", pyo.value(model.functionality))

            print("Functionality max epsilon (optimal value):", obj_3_max_epsilon)
            print("Economic Loss when optimizing Functionality:", obj_1_value_3)
            print("Dislocation when optimizing Functionality:", obj_2_value_3)
            results_list = [obj_3_max_epsilon, obj_1_value_3, obj_2_value_3]
        else:
            print("Not Optimal")
            results_list = [0, 0, 0]

        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time for initial obj 3 solve: ", elapsedtime)

        return results_list

    def configure_min_max_epsilon_values(self, model, objs, num_epsilon_steps):
        # Define positions in the incoming list
        __obj_1_value_2_pos = 4
        __obj_1_value_3_pos = 7
        __obj_1_min_epsilon_pos = 0
        __obj_2_value_1_pos = 1
        __obj_2_value_3_pos = 8
        __obj_2_min_epsilon_pos = 3
        __obj_3_value_1_pos = 2
        __obj_3_value_2_pos = 5
        __obj_3_max_epsilon_pos = 6
        model.econ_loss_max = Param(within=NonNegativeReals, initialize=max(objs[__obj_1_value_2_pos],
                                                                            objs[__obj_1_value_3_pos]))
        model.econ_loss_min = Param(within=NonNegativeReals, initialize=objs[__obj_1_min_epsilon_pos])
        model.dislocation_max = Param(within=NonNegativeReals, initialize=max(objs[__obj_2_value_1_pos],
                                                                              objs[__obj_2_value_3_pos]))
        model.dislocation_min = Param(within=NonNegativeReals, initialize=objs[__obj_2_min_epsilon_pos])
        model.functionality_max = Param(within=NonNegativeReals, initialize=objs[__obj_3_max_epsilon_pos])
        model.functionality_min = Param(within=NonNegativeReals, initialize=min(objs[__obj_3_value_1_pos],
                                                                                objs[__obj_3_value_2_pos]))

        model.econ_loss_step = Param(within=NonNegativeReals,
                                     initialize=(pyo.value(model.econ_loss_max) - pyo.value(model.econ_loss_min)) * (
                                                 1 / (num_epsilon_steps - 1)))
        model.dislocation_step = Param(within=NonNegativeReals, initialize=(pyo.value(
            model.dislocation_max) - pyo.value(model.dislocation_min)) * (1 / (num_epsilon_steps - 1)))
        model.functionality_step = Param(within=NonNegativeReals, initialize=(pyo.value(
            model.functionality_max) - pyo.value(model.functionality_min)) * (1 / (num_epsilon_steps - 1)))

    def solve_epsilon_models(self, model, model_solver_setting, inactive_submodels):
        xresults_df = pd.DataFrame()
        yresults_df = pd.DataFrame()

        if 1 not in inactive_submodels:
           self.solve_epsilon_model_1(model, model_solver_setting)

        if 2 not in inactive_submodels:
            self.solve_epsilon_model_2(model, model_solver_setting)

        if 3 not in inactive_submodels:
            self.solve_epsilon_model_3(model, model_solver_setting)

        if 4 not in inactive_submodels:
            self.solve_epsilon_model_4(model, model_solver_setting)

        if 5 not in inactive_submodels:
            self.solve_epsilon_model_5(model, model_solver_setting)

        if 6 not in inactive_submodels:
            self.solve_epsilon_model_6(model, model_solver_setting)

        if 7 not in inactive_submodels:
            xresults_df, yresults_df = self.solve_epsilon_model_7(model, model_solver_setting, xresults_df,
                                                                      yresults_df)

        if 8 not in inactive_submodels:
            xresults_df, yresults_df = self.solve_epsilon_model_8(model, model_solver_setting, xresults_df,
                                                                      yresults_df)

        if 9 not in inactive_submodels:
            xresults_df, yresults_df = self.solve_epsilon_model_9(model, model_solver_setting, xresults_df,
                                                                      yresults_df)

        return xresults_df, yresults_df

    def solve_epsilon_model_1(self, model, model_solver_setting):
        starttime = time.time()
        print("****OPTIMIZING ECONOMIC LOSS SUBJECT TO POPULATION DISLOCATION EPSILON CONSTRAINTS****")
        # Activate objective function 1 and deactivate others:
        model.objective_1.activate()
        model.objective_2.deactivate()
        model.objective_3.deactivate()

        # Add objective function 2 as epsilon constraint of objective function 1:
        # Dataframe to store optimization results:
        obj_1_2_epsilon_results = pd.DataFrame(
            columns=['Economic Loss(Million Dollars)', 'Dislocation Value', 'Functionality Value'])

        # Parameter for objective 2 (dislocation) epsilon value:
        model.obj_2_e = Param(mutable=True, within=NonNegativeReals)
        # For each dislocation epsilon value (starting at min) set as constraint and solve model:
        counter = 0
        for e in np.arange(pyo.value(model.dislocation_min), pyo.value(model.dislocation_max) + 0.000001,
                           pyo.value(model.dislocation_step)):
            counter += 1
            print("Step ", counter, ": ", e)

            model.obj_2_e = e  # Set the model parameter to the epsilon value.
            model.add_component("objective_2_constraint",
                                Constraint(expr=sum_product(model.d_ijk, model.x_ijk) <= pyo.value(
                                    model.obj_2_e)))  # Add the epsilon constraint.
            # Solve the model:
            results = model_solver_setting.solve(model)

            # Save the results if the solver returns an optimal solution:
            if (results.solver.status == SolverStatus.ok) and (
                    results.solver.termination_condition == TerminationCondition.optimal):
                self.extract_optimization_results(model)
                # Add objective (economic loss), dislocation, and functionality values to results dataframe:
                obj_1_2_epsilon_results.loc[counter - 1, 'Economic Loss(Million Dollars)'] = pyo.value(
                    model.econ_loss)  # Save the optimal economic loss value.
                obj_1_2_epsilon_results.loc[counter - 1, 'Dislocation Value'] = pyo.value(
                    model.dislocation)  # Save the resulting dislocation value.
                obj_1_2_epsilon_results.loc[counter - 1, 'Functionality Value'] = pyo.value(
                    model.functionality)  # Save the resulting functionality value.
            else:
                print(results.solver.termination_condition)
                log_infeasible_constraints(model)

            # Remove the given epsilon constraint from the model now that optimization is complete:
            model.del_component(model.objective_2_constraint)

        # Remove the dislocation epsilon parameter from the model:
        model.del_component(model.obj_2_e)

        # Display and save the results of optimizing economic loss subject to dislocation epsilon constraints:
        print(obj_1_2_epsilon_results)
        filename = 'obj_1_2_epsilon_results_' + str(time.strftime("%m-%d-%Y")) + '.csv'
        obj_1_2_epsilon_results.to_csv(filename)
        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time: ", elapsedtime)

    def solve_epsilon_model_2(self, model, model_solver_setting):
        starttime = time.time()
        print("****OPTIMIZING ECONOMIC LOSS SUBJECT TO BUILDING FUNCTIONALITY EPSILON CONSTRAINTS****")
        # Activate objective function 1 and deactivate others:
        model.objective_1.activate()
        model.objective_2.deactivate()
        model.objective_3.deactivate()

        # Add objective function 3 as epsilon constraint of objective function 1:
        # Dataframe to store optimization results:
        obj_1_3_epsilon_results = pd.DataFrame(
            columns=['Economic Loss(Million Dollars)', 'Dislocation Value', 'Functionality Value'])

        # Parameter for objective 3 (functionality) epsilon value:
        model.obj_3_e = Param(mutable=True, within=NonNegativeReals)
        counter = 0
        # For each functionality epsilon value (starting at min) set as constraint and solve model:
        # Adding 0.0000000000001 to the maximum value allows np.arange() to include the maximum functionality.
        for e in np.arange(pyo.value(model.functionality_min),
                           pyo.value(model.functionality_max) + 0.000000000000000001,
                           pyo.value(model.functionality_step)):
            counter += 1
            print("Step ", counter, ": ", e)

            model.obj_3_e = e  # Set the model parameter to the epsilon value.
            model.add_component("objective_3_constraint",
                                Constraint(expr=sum_product(model.Q_t_hat, model.x_ijk) >= pyo.value(
                                    model.obj_3_e)))  # Add the epsilon constraint.
            # Solve the model:
            results = model_solver_setting.solve(model)

            # Save the results if the solver returns an optimal solution:
            if (results.solver.status == SolverStatus.ok) and (
                    results.solver.termination_condition == TerminationCondition.optimal):
                self.extract_optimization_results(model)
                # Add objective (economic loss), dislocation, and functionality values to results dataframe:
                obj_1_3_epsilon_results.loc[counter - 1, 'Economic Loss(Million Dollars)'] = pyo.value(
                    model.econ_loss)  # Save the optimal economic loss value.
                obj_1_3_epsilon_results.loc[counter - 1, 'Dislocation Value'] = pyo.value(
                    model.dislocation)  # Save the resulting dislocation value.
                obj_1_3_epsilon_results.loc[counter - 1, 'Functionality Value'] = pyo.value(
                    model.functionality)  # Save the resulting functionality value.
            else:
                print(results.solver.termination_condition)
                log_infeasible_constraints(model)

            # Remove the given epsilon constraint from the model now that optimization is complete:
            model.del_component(model.objective_3_constraint)

        # Remove the functionality epsilon parameter from the model:
        model.del_component(model.obj_3_e)

        # Display and save the results of optimizing economic loss subject to functionality epsilon constraints:
        print(obj_1_3_epsilon_results)
        filename = 'obj_1_3_epsilon_results_' + str(time.strftime("%m-%d-%Y")) + '.csv'
        obj_1_3_epsilon_results.to_csv(filename)
        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time: ", elapsedtime)

    def solve_epsilon_model_3(self, model, model_solver_setting):
        starttime = time.time()
        print("****OPTIMIZING POPULATION DISLOCATION SUBJECT TO ECONOMIC LOSS EPSILON CONSTRAINTS****")
        # Activate objective function 2 and deactivate others:
        model.objective_1.deactivate()
        model.objective_2.activate()
        model.objective_3.deactivate()

        # Add objective function 1 as epsilon constraint of objective function 2:
        # Dataframe to store optimization results:
        obj_2_1_epsilon_results = pd.DataFrame(
            columns=['Economic Loss(Million Dollars)', 'Dislocation Value', 'Functionality Value'])

        # Parameter for objective 1 (economic loss) epsilon value:
        model.obj_1_e = Param(mutable=True, within=NonNegativeReals)
        # For each economic loss epsilon value (starting at min) set as constraint and solve model:
        # Adding 1 to the maximum value allows np.arange() to include the maximum economic loss.
        counter = 0
        for e in np.arange(pyo.value(model.econ_loss_min), pyo.value(model.econ_loss_max) + 0.1,
                           pyo.value(model.econ_loss_step)):
            counter += 1
            print("Step ", counter, ": ", e)

            model.obj_1_e = e  # Set the model parameter to the epsilon value.
            model.add_component("objective_1_constraint",
                                Constraint(expr=sum_product(model.l_ijk, model.x_ijk) <= pyo.value(
                                    model.obj_1_e)))  # Set epsilon constraint.
            # Solve the model:
            results = model_solver_setting.solve(model)

            # Save the results if the solver returns an optimal solution:
            if (results.solver.status == SolverStatus.ok) and (
                    results.solver.termination_condition == TerminationCondition.optimal):
                self.extract_optimization_results(model)
                # Add objective (dislocation), economic loss, and functionality values to results dataframe:
                obj_2_1_epsilon_results.loc[counter - 1, 'Dislocation Value'] = pyo.value(
                    model.dislocation)  # Save the optimal dislocation value.
                obj_2_1_epsilon_results.loc[counter - 1, 'Economic Loss(Million Dollars)'] = pyo.value(
                    model.econ_loss)  # Save the resulting economic loss value.
                obj_2_1_epsilon_results.loc[counter - 1, 'Functionality Value'] = pyo.value(
                    model.functionality)  # Save the resulting functionality value.
            else:
                print(results.solver.termination_condition)
                log_infeasible_constraints(model)

            # Remove the given epsilon constraint from the model now that optimization is complete:
            model.del_component(model.objective_1_constraint)

        # Remove the functionality epsilon parameter from the model:
        model.del_component(model.obj_1_e)

        # Display the results of optimizing dislocation subject to economic loss epsilon constraints:
        print(obj_2_1_epsilon_results)
        filename = 'obj_2_1_epsilon_results_' + str(time.strftime("%m-%d-%Y")) + '.csv'
        obj_2_1_epsilon_results.to_csv(filename)
        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time: ", elapsedtime)

    def solve_epsilon_model_4(self, model, model_solver_setting):
        starttime = time.time()
        print("****OPTIMIZING POPULATION DISLOCATION SUBJECT TO BUILDING FUNCTIONALITY EPSILON CONSTRAINTS****")
        # Activate objective function 2 and deactivate others:
        model.objective_1.deactivate()
        model.objective_2.activate()
        model.objective_3.deactivate()

        # Add objective function 3 as epsilon constraint of objective function 2:
        # Dataframe to store optimization results:
        obj_2_3_epsilon_results = pd.DataFrame(
            columns=['Economic Loss(Million Dollars)', 'Dislocation Value', 'Functionality Value'])

        # Parameter for objective 3 (functionality) epsilon value:
        model.obj_3_e = Param(mutable=True, within=NonNegativeReals)
        # For each functionality epsilon value (starting at min) set as constraint and solve model:
        # Adding 0.0000000000001 to the maximum value allows np.arange() to include the maximum functionality.
        counter = 0
        for e in np.arange(pyo.value(model.functionality_min), pyo.value(model.functionality_max) + 0.0000000000001,
                           pyo.value(model.functionality_step)):
            counter += 1
            print("Step ", counter, ": ", e)

            model.obj_3_e = e  # Set the model parameter to the epsilon value.
            model.add_component("objective_3_constraint",
                                Constraint(expr=sum_product(model.Q_t_hat, model.x_ijk) >= (
                                    pyo.value(model.obj_3_e))))  # Set the epsilon constraint.
            # Solve the model:
            results = model_solver_setting.solve(model)

            # Save the results if the solver returns an optimal solution:
            if (results.solver.status == SolverStatus.ok) and (
                    results.solver.termination_condition == TerminationCondition.optimal):
                self.extract_optimization_results(model)
                # Add objective (dislocation), economic loss, and functionality values to results dataframe:
                obj_2_3_epsilon_results.loc[counter - 1, 'Dislocation Value'] = pyo.value(
                    model.dislocation)  # Save the optimal dislocation value.
                obj_2_3_epsilon_results.loc[counter - 1, 'Economic Loss(Million Dollars)'] = pyo.value(
                    model.econ_loss)  # Save the resulting economic loss value.
                obj_2_3_epsilon_results.loc[counter - 1, 'Functionality Value'] = pyo.value(
                    model.functionality)  # Save the resulting functionality value.
            else:
                print(results.solver.termination_condition)
                log_infeasible_constraints(model)

            # Remove the given epsilon constraint from the model now that optimization is complete:
            model.del_component(model.objective_3_constraint)

        # Remove the functionality epsilon parameter from the model:
        model.del_component(model.obj_3_e)

        # Display the results of optimizing dislocation subject to functionality epsilon constraints:
        print(obj_2_3_epsilon_results)
        filename = 'obj_2_3_epsilon_results_' + str(time.strftime("%m-%d-%Y")) + '.csv'
        obj_2_3_epsilon_results.to_csv(filename)
        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time: ", elapsedtime)

    def solve_epsilon_model_5(self, model, model_solver_setting):
        starttime = time.time()
        print("****OPTIMIZING BUILDING FUNCTIONALITY SUBJECT TO ECONOMIC LOSS EPSILON CONSTRAINTS****")
        # Activate objective function 3 and deactivate others:
        model.objective_1.deactivate()
        model.objective_2.deactivate()
        model.objective_3.activate()

        # Add objective function 1 as epsilon constraint of objective function 3:
        # Dataframe to store optimization results:
        obj_3_1_epsilon_results = pd.DataFrame(
            columns=['Economic Loss(Million Dollars)', 'Dislocation Value', 'Functionality Value'])

        # Parameter for objective 1 (economic loss) epsilon value:
        model.obj_1_e = Param(mutable=True, within=NonNegativeReals)
        # For each economic loss epsilon value (starting at min) set as constraint and solve model:
        # Adding 1 to the maximum value allows np.arange() to include the maximum economic loss.
        counter = 0
        for e in np.arange(pyo.value(model.econ_loss_min), pyo.value(model.econ_loss_max) + 0.000001,
                           pyo.value(model.econ_loss_step)):
            counter += 1
            print("Step ", counter, ": ", e)

            model.obj_1_e = e
            model.add_component("objective_1_constraint",
                                Constraint(expr=sum_product(model.l_ijk, model.x_ijk) <= pyo.value(
                                    model.obj_1_e)))  # Set the epsilon constraint.
            # Solve the model:
            results = model_solver_setting.solve(model)

            # Save the results if the solver returns an optimal solution:
            if (results.solver.status == SolverStatus.ok) and (
                    results.solver.termination_condition == TerminationCondition.optimal):
                self.extract_optimization_results(model)
                # Add objective (functionality), economic loss, and dislocation values to results dataframe:
                obj_3_1_epsilon_results.loc[counter - 1, 'Functionality Value'] = pyo.value(
                    model.functionality)  # Save the optimal functionality value.
                obj_3_1_epsilon_results.loc[counter - 1, 'Economic Loss(Million Dollars)'] = pyo.value(
                    model.econ_loss)  # Save the resulting economic loss value.
                obj_3_1_epsilon_results.loc[counter - 1, 'Dislocation Value'] = pyo.value(
                    model.dislocation)  # Save the resulting dislocation value.
            else:
                print(results.solver.termination_condition)
                log_infeasible_constraints(model)

            # Remove the given epsilon constraint from the model now that optimization is complete:
            model.del_component(model.objective_1_constraint)

        # Remove the economic loss epsilon parameter from the model:
        model.del_component(model.obj_1_e)

        # Display the results of optimizing functionality subject to economic loss epsilon constraints:
        print(obj_3_1_epsilon_results)
        filename = 'obj_3_1_epsilon_results_' + str(time.strftime("%m-%d-%Y")) + '.csv'
        obj_3_1_epsilon_results.to_csv(filename)
        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time: ", elapsedtime)

    def solve_epsilon_model_6(self, model, model_solver_setting):
        starttime = time.time()
        print("****OPTIMIZING BUILDING FUNCTIONALITY SUBJECT TO POPULATION DISLOCATION EPSILON CONSTRAINTS****")
        # Activate objective function 3 (functionality) and deactivate others:
        model.objective_1.deactivate()
        model.objective_2.deactivate()
        model.objective_3.activate()

        # Add objective function 2 as epsilon constraint of objective function 3:
        # Dataframe to store optimization results:
        obj_3_2_epsilon_results = pd.DataFrame(
            columns=['Economic Loss(Million Dollars)', 'Dislocation Value', 'Functionality Value'])

        # Parameter for objective 2 (dislocation) epsilon value:
        model.obj_2_e = Param(mutable=True, within=NonNegativeReals)
        # For each dislocation epsilon value (starting at min) set as constraint and solve model:
        # Adding 1 to the maximum value allows np.arange() to include the maximum dislocation.
        counter = 0
        for e in np.arange(pyo.value(model.dislocation_min), pyo.value(model.dislocation_max) + 0.000001,
                           pyo.value(model.dislocation_step)):
            counter += 1
            print("Step ", counter, ": ", e)

            model.obj_2_e = e
            model.add_component("objective_2_constraint",
                                Constraint(expr=sum_product(model.d_ijk, model.x_ijk) <= pyo.value(
                                    model.obj_2_e)))  # Set the epsilon constraint.
            # Solve the model:
            results = model_solver_setting.solve(model)

            # Save the results if the solver returns an optimal solution:
            if (results.solver.status == SolverStatus.ok) and (
                    results.solver.termination_condition == TerminationCondition.optimal):
                self.extract_optimization_results(model)
                # Add objective (functionality), economic loss, and dislocation values to results dataframe:
                obj_3_2_epsilon_results.loc[counter - 1, 'Functionality Value'] = pyo.value(
                    model.functionality)  # Save the optimal functionality value.
                obj_3_2_epsilon_results.loc[counter - 1, 'Economic Loss(Million Dollars)'] = pyo.value(
                    model.econ_loss)  # Save the resulting economic loss value.
                obj_3_2_epsilon_results.loc[counter - 1, 'Dislocation Value'] = pyo.value(
                    model.dislocation)  # Save the resulting dislocation value.
            else:
                print(results.solver.termination_condition)
                log_infeasible_constraints(model)

            # Remove the given epsilon constraint from the model now that optimization is complete:
            model.del_component(model.objective_2_constraint)

        # Remove the economic loss epsilon parameter from the model:
        model.del_component(model.obj_2_e)

        # Display the results of optimizing functionality subject to dislocation epsilon constraints:
        print(obj_3_2_epsilon_results)
        filename = 'obj_3_2_epsilon_results_' + str(time.strftime("%m-%d-%Y")) + '.csv'
        obj_3_2_epsilon_results.to_csv(filename)
        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time: ", elapsedtime)

    def solve_epsilon_model_7(self, model, model_solver_setting, xresults_df, yresults_df):
        starttime = time.time()
        print(
            "****OPTIMIZING ECONOMIC LOSS SUBJECT TO POPULATION DISLOCATION "
            "AND BUILDING FUNCTIONALITY EPSILON CONSTRAINTS****")
        model.objective_1.activate()
        model.objective_2.deactivate()
        model.objective_3.deactivate()

        # Add objective functions 2 and 3 as epsilon constraints of objective function 1:
        # Dataframe to store optimization results:
        obj_1_23_epsilon_results = pd.DataFrame(
            columns=['Economic Loss(Million Dollars)', 'Dislocation Value', 'Functionality Value'])

        # Parameter for objective 2 (dislocation) and objective 3 (functionality) epsilon values:
        model.obj_2_e = Param(mutable=True, within=NonNegativeReals)
        model.obj_3_e = Param(mutable=True, within=NonNegativeReals)

        # Generate datasets for this epsilon analysis
        epsilon7_xresult_df = pd.DataFrame()
        epsilon7_yresult_df = pd.DataFrame()

        counter = 0
        for e in np.arange(pyo.value(model.dislocation_min), pyo.value(model.dislocation_max) + 0.000001,
                           pyo.value(model.dislocation_step)):
            model.obj_2_e = e
            model.add_component("objective_2_constraint",
                                Constraint(expr=sum_product(model.d_ijk, model.x_ijk) <= pyo.value(model.obj_2_e)))
            for e2 in np.arange(pyo.value(model.functionality_min),
                                pyo.value(model.functionality_max) + 0.0000000000001,
                                pyo.value(model.functionality_step)):
                counter += 1
                model.obj_3_e = e2
                print("Step ", counter, " e: ", e, "  e2:", e2)
                #             obj_1_23_epsilon_results.loc[counter-1,'Dislocation Epsilon']=e
                #             obj_1_23_epsilon_results.loc[counter-1,'Functionality Epsilon']=e2

                model.add_component("objective_3_constraint",
                                    Constraint(
                                        expr=sum_product(model.Q_t_hat, model.x_ijk) >= pyo.value(model.obj_3_e)))
                # Solve the model:
                results = model_solver_setting.solve(model)

                # Save the results if the solver returns an optimal solution:
                if (results.solver.status == SolverStatus.ok) and (
                        results.solver.termination_condition == TerminationCondition.optimal):
                    self.extract_optimization_results(model)
                    budget_used = quicksum(pyo.value(model.y_ijkk_prime[i, j, k, k_prime]) * pyo.value(
                        model.Sc_ijkk_prime[i, j, k, k_prime]) for (i, j, k, k_prime) in model.ZSKK_prime)
                    percent_budget_used = (budget_used / pyo.value(
                        model.B)) * 100  # Record percentage of available budget used.
                    # Add objective (economic loss), dislocation, and functionality values to results dataframe:
                    # TODO: optimize this search
                    obj_1_23_epsilon_results.loc[counter - 1, 'Functionality Value'] = pyo.value(model.functionality)
                    obj_1_23_epsilon_results.loc[counter - 1, 'Economic Loss(Million Dollars)'] = pyo.value(
                        model.econ_loss)
                    obj_1_23_epsilon_results.loc[counter - 1, 'Dislocation Value'] = pyo.value(model.dislocation)

                    # Extract results per each variable and convert to dataframe
                    x_data: dict = model.x_ijk.extract_values()
                    newx_df = self.assemble_dataframe_from_solution("x_ijk", x_data, counter)

                    y_data: dict = model.y_ijkk_prime.extract_values()
                    newy_df = self.assemble_dataframe_from_solution("y_ijkk_prime", y_data, counter)

                    # Append to local analysis result
                    epsilon7_xresult_df = pd.concat([epsilon7_xresult_df, newx_df])
                    epsilon7_yresult_df = pd.concat([epsilon7_yresult_df, newy_df])
                else:
                    print(results.solver.termination_condition)
                    log_infeasible_constraints(model)

                # Remove the given epsilon constraint from the model now that optimization is complete:
                model.del_component(model.objective_3_constraint)

            # Remove the given epsilon constraint from the model now that optimization is complete:
            model.del_component(model.objective_2_constraint)

        # Remove the economic loss epsilon parameter from the model:
        model.del_component(model.obj_2_e)
        model.del_component(model.obj_3_e)

        print(obj_1_23_epsilon_results)

        # Drop rows with infeasible results:
        initial_length = len(obj_1_23_epsilon_results)
        obj_1_23_data = obj_1_23_epsilon_results.dropna(axis=0, how='any')
        print("Infeasible rows dropped: ", initial_length - len(obj_1_23_data), " rows.")

        # Save results:
        filename = '7_epsilon_results.csv'
        obj_1_23_data.index += 1
        obj_1_23_data.to_csv(filename)

        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time: ", elapsedtime)

        epsilon7_xresult_df['Epsilon'] = 7
        epsilon7_yresult_df['Epsilon'] = 7

        return pd.concat([xresults_df, epsilon7_xresult_df]), pd.concat([yresults_df, epsilon7_yresult_df])

    def solve_epsilon_model_8(self, model, model_solver_setting, xresults_df, yresults_df):
        starttime = time.time()
        print(
            "****OPTIMIZING POPULATION DISLOCATION SUBJECT TO "
            "ECONOMIC LOSS AND BUILDING FUNCTIONALITY EPSILON CONSTRAINTS****")
        model.objective_1.deactivate()
        model.objective_2.activate()
        model.objective_3.deactivate()

        # Add objective functions 1 and 3 as epsilon constraints of objective function 2:
        # Dataframe to store optimization results:
        obj_2_13_epsilon_results = pd.DataFrame(
            columns=['Economic Loss(Million Dollars)', 'Dislocation Value', 'Functionality Value'])

        # Parameter for objective 1 (economic loss) and objective 3 (functionality) epsilon values:
        model.obj_1_e = Param(mutable=True, within=NonNegativeReals)
        model.obj_3_e = Param(mutable=True, within=NonNegativeReals)

        # Generate datasets for this epsilon analysis
        epsilon8_xresult_df = pd.DataFrame()
        epsilon8_yresult_df = pd.DataFrame()

        counter = 0
        for e in np.arange(pyo.value(model.econ_loss_min), pyo.value(model.econ_loss_max) + 0.000001,
                           pyo.value(model.econ_loss_step)):
            model.obj_1_e = e
            model.add_component("objective_1_constraint",
                                Constraint(expr=sum_product(model.l_ijk, model.x_ijk) <= pyo.value(model.obj_1_e)))
            for e2 in np.arange(pyo.value(model.functionality_min), pyo.value(model.functionality_max) +
                                0.0000000000001, pyo.value(model.functionality_step)):
                counter += 1
                model.obj_3_e = e2
                print("Step ", counter, " e: ", e, "  e2:", e2)
                #             obj_2_13_epsilon_results.loc[counter-1,'Economic Loss Epsilon']=e
                #             obj_2_13_epsilon_results.loc[counter-1,'Functionality Epsilon']=e2
                model.add_component("objective_3_constraint",
                                    Constraint(
                                        expr=sum_product(model.Q_t_hat, model.x_ijk) >= pyo.value(model.obj_3_e)))
                # Solve the model:
                results = model_solver_setting.solve(model)

                # Save the results if the solver returns an optimal solution:
                if (results.solver.status == SolverStatus.ok) and (
                        results.solver.termination_condition == TerminationCondition.optimal):
                    self.extract_optimization_results(model)
                    budget_used = quicksum(pyo.value(model.y_ijkk_prime[i, j, k, k_prime]) * pyo.value(
                        model.Sc_ijkk_prime[i, j, k, k_prime]) for (i, j, k, k_prime) in model.ZSKK_prime)
                    percent_budget_used = (budget_used / pyo.value(
                        model.B)) * 100  # Record percentage of available budget used.
                    # Add objective (dislocation), economic loss, and functionality values to results dataframe:
                    # TODO: optimize this search
                    obj_2_13_epsilon_results.loc[counter - 1, 'Functionality Value'] = pyo.value(model.functionality)
                    obj_2_13_epsilon_results.loc[counter - 1, 'Economic Loss(Million Dollars)'] = pyo.value(
                        model.econ_loss)
                    obj_2_13_epsilon_results.loc[counter - 1, 'Dislocation Value'] = pyo.value(model.dislocation)

                    # Extract results per each variable and convert to dataframe
                    x_data: dict = model.x_ijk.extract_values()
                    newx_df = self.assemble_dataframe_from_solution("x_ijk", x_data, counter)

                    y_data: dict = model.y_ijkk_prime.extract_values()
                    newy_df = self.assemble_dataframe_from_solution("y_ijkk_prime", y_data, counter)

                    # Append to local analysis result
                    epsilon8_xresult_df = pd.concat([epsilon8_xresult_df, newx_df])
                    epsilon8_yresult_df = pd.concat([epsilon8_yresult_df, newy_df])
                else:
                    print(results.solver.termination_condition)
                    log_infeasible_constraints(model)

                # Remove the given epsilon constraint from the model now that optimization is complete:
                model.del_component(model.objective_3_constraint)

            # Remove the given epsilon constraint from the model now that optimization is complete:
            model.del_component(model.objective_1_constraint)

        # Remove the economic loss epsilon parameter from the model:
        model.del_component(model.obj_1_e)
        model.del_component(model.obj_3_e)

        print(obj_2_13_epsilon_results)

        # Drop rows with infeasible results:
        initial_length = len(obj_2_13_epsilon_results)
        obj_2_13_data = obj_2_13_epsilon_results.dropna(axis=0, how='any')
        print("Infeasible rows dropped: ", initial_length - len(obj_2_13_data), " rows.")

        # Save results:
        filename = '8_epsilon_results.csv'
        obj_2_13_data.index += 1
        obj_2_13_data.to_csv(filename)

        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time: ", elapsedtime)

        epsilon8_xresult_df['Epsilon'] = 8
        epsilon8_yresult_df['Epsilon'] = 8

        return pd.concat([xresults_df, epsilon8_xresult_df]), pd.concat([yresults_df, epsilon8_yresult_df])

    def solve_epsilon_model_9(self, model, model_solver_setting, xresults_df, yresults_df):
        starttime = time.time()
        print(
            "****OPTIMIZING BUILDING FUNCTIONALITY SUBJECT TO "
            "ECONOMIC LOSS AND POPULATION DISLOCATION EPSILON CONSTRAINTS****")
        model.objective_1.deactivate()
        model.objective_2.deactivate()
        model.objective_3.activate()

        # Add objective functions 1 and 2 as epsilon constraints of objective function 3:
        # Dataframe to store optimization results:
        obj_3_12_epsilon_results = pd.DataFrame(
            columns=['Economic Loss(Million Dollars)', 'Dislocation Value', 'Functionality Value'])

        # Parameter for objective 2 (dislocation) and objective 1 (economic loss) epsilon values:
        model.obj_1_e = Param(mutable=True, within=NonNegativeReals)
        model.obj_2_e = Param(mutable=True, within=NonNegativeReals)

        # Generate datasets for this epsilon analysis
        epsilon9_xresult_df = pd.DataFrame()
        epsilon9_yresult_df = pd.DataFrame()

        counter = 0
        for e in np.arange(pyo.value(model.econ_loss_min), pyo.value(model.econ_loss_max),
                           pyo.value(model.econ_loss_step)):
            model.obj_1_e = e
            model.add_component("objective_1_constraint",
                                Constraint(expr=sum_product(model.l_ijk, model.x_ijk) <= pyo.value(model.obj_1_e)))
            for e2 in np.arange(pyo.value(model.dislocation_min), pyo.value(model.dislocation_max),
                                pyo.value(model.dislocation_step)):
                counter += 1
                model.obj_2_e = e2
                print("Step ", counter, " e: ", e, "  e2:", e2)
                #             obj_3_12_epsilon_results.loc[counter-1,'Economic Loss Epsilon']=e
                #             obj_3_12_epsilon_results.loc[counter-1,'Dislocation Epsilon']=e2
                model.add_component("objective_2_constraint",
                                    Constraint(
                                        expr=sum_product(model.d_ijk, model.x_ijk) <= (pyo.value(model.obj_2_e))))
                # Solve the model:
                results = model_solver_setting.solve(model)

                # Save the results if the solver returns an optimal solution:
                if (results.solver.status == SolverStatus.ok) and (
                        results.solver.termination_condition == TerminationCondition.optimal):
                    self.extract_optimization_results(model)
                    budget_used = quicksum(pyo.value(model.y_ijkk_prime[i, j, k, k_prime]) * pyo.value(
                        model.Sc_ijkk_prime[i, j, k, k_prime]) for (i, j, k, k_prime) in model.ZSKK_prime)
                    percent_budget_used = (budget_used / pyo.value(
                        model.B)) * 100  # Record percentage of available budget used.
                    # Add objective (functionality), economic loss, and dislocation values to results dataframe:
                    # TODO: optimize this search
                    obj_3_12_epsilon_results.loc[counter - 1, 'Functionality Value'] = pyo.value(model.functionality)
                    obj_3_12_epsilon_results.loc[counter - 1, 'Economic Loss(Million Dollars)'] = pyo.value(
                        model.econ_loss)
                    obj_3_12_epsilon_results.loc[counter - 1, 'Dislocation Value'] = pyo.value(model.dislocation)

                    # Extract results per each variable and convert to dataframe
                    x_data: dict = model.x_ijk.extract_values()
                    newx_df = self.assemble_dataframe_from_solution("x_ijk", x_data, counter)

                    y_data: dict = model.y_ijkk_prime.extract_values()
                    newy_df = self.assemble_dataframe_from_solution("y_ijkk_prime", y_data, counter)

                    # Append to local analysis result
                    epsilon9_xresult_df = pd.concat([epsilon9_xresult_df, newx_df])
                    epsilon9_yresult_df = pd.concat([epsilon9_yresult_df, newy_df])
                else:
                    print(results.solver.termination_condition)
                    log_infeasible_constraints(model)

                # Remove the given epsilon constraint from the model now that optimization is complete:
                model.del_component(model.objective_2_constraint)

            # Remove the given epsilon constraint from the model now that optimization is complete:
            model.del_component(model.objective_1_constraint)

        # Remove the economic loss epsilon parameter from the model:
        model.del_component(model.obj_1_e)
        model.del_component(model.obj_2_e)

        print(obj_3_12_epsilon_results)

        # Drop rows with infeasible results:
        initial_length = len(obj_3_12_epsilon_results)
        obj_3_12_data = obj_3_12_epsilon_results.dropna(axis=0, how='any')
        print("Infeasible rows dropped: ", initial_length - len(obj_3_12_data), " rows.")

        # Save results:
        filename = '9_epsilon_results.csv'
        obj_3_12_data.index += 1
        obj_3_12_data.to_csv(filename)

        endtime = time.time()
        elapsedtime = endtime - starttime
        print("Elapsed time: ", elapsedtime)

        epsilon9_xresult_df['Epsilon'] = 9
        epsilon9_yresult_df['Epsilon'] = 9

        return pd.concat([xresults_df, epsilon9_xresult_df]), pd.concat([yresults_df, epsilon9_yresult_df])

    def compute_optimal_results(self, inactive_submodels, xresults_df, yresults_df):
        # Fixed for the moment, will be expanded to the full model set in later iterations
        epsilon_models = [7, 8, 9]

        xresults_list: list = []
        yresults_list: list = []

        for k in epsilon_models:
            if k not in inactive_submodels:
                results = pd.read_csv(str(k) + '_epsilon_results.csv', usecols=['Economic Loss(Million Dollars)',
                                                                             'Dislocation Value',
                                                                             'Functionality Value'],
                                      low_memory=False)
                list_loss = results['Economic Loss(Million Dollars)'].values.tolist()
                list_dislocation = results['Dislocation Value'].values.tolist()
                list_func = results['Functionality Value'].values.tolist()
                zipped_list = self.optimal_points(list_loss, list_dislocation, list_func)
                optimal = pd.DataFrame(zipped_list, columns=['Iteration', 'Economic Loss(Million Dollars)',
                                                             'Dislocation Value', 'Functionality Value'])

                # Select only results corresponding to current epsilon step
                epsilon_xresults_df = xresults_df[xresults_df['Epsilon'] == k]
                epsilon_yresults_df = yresults_df[yresults_df['Epsilon'] == k]

                # Filter results depending on optimal epsilon model
                opt_xresults_df = epsilon_xresults_df[epsilon_xresults_df['Iteration'].isin(optimal['Iteration'])]
                opt_yresults_df = epsilon_yresults_df[epsilon_yresults_df['Iteration'].isin(optimal['Iteration'])]

                # Append for later construction of final dataset
                xresults_list.append(opt_xresults_df)
                yresults_list.append(opt_yresults_df)

        # Construct the final dataframe per variable
        return [pd.concat(xresults_list), pd.concat(yresults_list)]

    # Objective functions
    @staticmethod
    def obj_economic(model):
        # return(sum_product(model.l_ijk,model.x_ijk))
        return quicksum(model.l_ijk[i, j, k] * model.x_ijk[i, j, k] for (i, j, k) in model.ZSK)

    @staticmethod
    def obj_dislocation(model):
        # return(sum_product(model.d_ijk,model.x_ijk))
        return quicksum(model.d_ijk[i, j, k] * model.x_ijk[i, j, k] for (i, j, k) in model.ZSK)

    @staticmethod
    def obj_functionality(model):
        # return(sum_product(model.Q_t_hat,model.x_ijk))
        return quicksum(model.Q_t_hat[i, j, k] * model.x_ijk[i, j, k] for (i, j, k) in model.ZSK)

    @staticmethod
    def retrofit_cost_rule(model):
        return (None,
                quicksum(
                    model.Sc_ijkk_prime[i, j, k, k_prime] * model.y_ijkk_prime[i, j, k, k_prime] for
                    (i, j, k, k_prime)
                    in model.ZSKK_prime),
                pyo.value(model.B))

    @staticmethod
    def number_buildings_ij_rule(model, i, j):
        return (quicksum(pyo.value(model.b_ijk[i, j, k]) for k in model.K),
                quicksum(model.x_ijk[i, j, k] for k in model.K),
                quicksum(pyo.value(model.b_ijk[i, j, k]) for k in model.K))

    @staticmethod
    def building_level_rule(model, i, j, k):
        model.a = quicksum(model.y_ijkk_prime[i, j, k_prime, k] for k_prime in model.K_prime if
                           (i, j, k_prime, k) in model.zskk_prime)
        model.c = quicksum(model.y_ijkk_prime[i, j, k, k_prime] for k_prime in model.K_prime if
                           (i, j, k, k_prime) in model.zskk_prime)
        return (pyo.value(model.b_ijk[i, j, k]),
                model.x_ijk[i, j, k] + quicksum(model.y_ijkk_prime[i, j, k, k_prime] for k_prime in model.K_prime if
                                                (i, j, k, k_prime) in model.zskk_prime) -
                quicksum(model.y_ijkk_prime[i, j, k_prime, k] for k_prime in model.K_prime if
                         (i, j, k_prime, k) in model.zskk_prime),
                pyo.value(model.b_ijk[i, j, k]))

    @staticmethod
    def extract_optimization_results(model):
        model.econ_loss = quicksum(
            pyo.value(model.l_ijk[i, j, k]) * pyo.value(model.x_ijk[i, j, k]) for (i, j, k) in model.ZSK)
        model.dislocation = quicksum(
            pyo.value(model.d_ijk[i, j, k]) * pyo.value(model.x_ijk[i, j, k]) for (i, j, k) in model.ZSK)
        model.functionality = quicksum(
            pyo.value(model.Q_t_hat[i, j, k]) * pyo.value(model.x_ijk[i, j, k]) for (i, j, k) in model.ZSK)

    @staticmethod
    def assemble_dataframe_from_solution(variable, sol_dict, iteration):
        x_index_lists = list(map(list, zip(*sol_dict.keys())))
        x_dict = {
            "Z": x_index_lists[0],
            "S": x_index_lists[1],
            "K": x_index_lists[2],
        }

        if variable == "y_ijkk_prime":
            x_dict["K'"] = x_index_lists[3]

        x_dict[variable] = sol_dict.values()

        df = pd.DataFrame(x_dict)
        df['Iteration'] = iteration

        return df

    @staticmethod
    def optimal_points(list_loss, list_dislocation, list_func):
        l_com = []
        for i in range(0, len(list_loss)):
            l_com.append(tuple([list_loss[i], list_dislocation[i], list_func[i]]))
        # remove the repeated the points
        set_com = set(l_com)

        # change the set to list
        list_temp = []
        for x in set_com:
            list_temp.append(x)

        final_f = []
        f_temp = []
        for i in range(len(list_temp)):
            for j in range(i + 1, len(list_temp)):
                if list_temp[i][0] == list_temp[j][0] and list_temp[i][1] == list_temp[j][1]:
                    if list_temp[i][2] > list_temp[j][2]:
                        f_temp.append(list_temp[j])

                    elif list_temp[j][2] > list_temp[i][2]:
                        f_temp.append(list_temp[i])
        final_f = list(set(list_temp) - set(f_temp))

        # find the optimal value for dislocation if loss and functionality are same as the last loop
        final_d = []
        d_temp = []
        for i in range(len(final_f)):
            for j in range(i + 1, len(final_f)):
                if final_f[i][0] == final_f[j][0] and final_f[i][2] == final_f[j][2]:
                    if final_f[i][1] < final_f[j][1]:
                        d_temp.append(final_f[j])
                    elif final_f[i][1] > final_f[j][1]:
                        d_temp.append(final_f[i])

        final_d = list(set(final_f) - set(d_temp))

        # find the optimal value for loss if dislocation and functionality are same as the last loop
        final_l = []
        l_temp = []
        for i in range(len(final_d)):
            for j in range(i + 1, len(final_d)):
                if final_d[i][1] == final_d[j][1] and final_d[i][2] == final_d[j][2]:
                    if final_d[i][0] < final_d[j][0]:
                        l_temp.append(final_d[j])
                    elif final_d[i][0] > final_d[j][0]:
                        l_temp.append(final_d[i])
        final_l = list(set(final_d) - set(l_temp))

        loss_optimal = []
        dislocation_optimal = []
        func_optimal = []
        for x in final_l:
            loss_optimal.append(x[0])
            dislocation_optimal.append(x[1])
            func_optimal.append(x[2])
        comb1 = []
        for i in range(len(list_loss)):
            comb1.append([list_loss[i], list_dislocation[i], list_func[i]])

        comb2 = []
        for i in range(len(loss_optimal)):
            comb2.append([loss_optimal[i], dislocation_optimal[i],
                          func_optimal[i]])

        optimal_index = {}
        for i in range(len(comb2)):
            if comb2[i] not in comb1:
                pass
            else:
                optimal_index[comb1.index(comb2[i])] = comb2[i]
        optimal_index_list = list(optimal_index.keys())
        zipped_list = list(zip(optimal_index_list, loss_optimal, dislocation_optimal, func_optimal))
        return zipped_list

    def get_spec(self):
        """Get specifications of the multiobjective retrofit optimization model.

        Returns:
            obj: A JSON object of specifications of the multiobjective retrofit optimization model.

        """
        return {
            'name': 'multiobjective-retrofit-optimization',
            'description': 'Multiobjective retrofit optimization model',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': False,
                    'description': 'Result CSV dataset name',
                    'type': str
                },
                {
                    'id': 'model_solver',
                    'required': False,
                    'description': 'Choice of the model solver to use. Gurobi is the default solver.',
                    'type': str
                },
                {
                    'id': 'num_epsilon_steps',
                    'required': True,
                    'description': 'Number of epsilon values to evaluate',
                    'type': int
                },
                {
                    'id': 'max_budget',
                    'required': True,
                    'description': 'Selection of maximum possible budget',
                    'type': str
                },
                {
                    'id': 'budget_available',
                    'required': False,
                    'description': 'Custom budget value',
                    'type': float
                },
                {
                    'id': 'inactive_submodels',
                    'required': False,
                    'description': 'Identifier of submodels to inactivate during analysis',
                    'type': List[int]
                },
                {
                    'id': 'scale_data',
                    'required': True,
                    'description': 'Choice for scaling data',
                    'type': bool
                },
                {
                    'id': 'scaling_factor',
                    'required': False,
                    'description': 'Custom scaling factor',
                    'type': float
                },
            ],
            'input_datasets': [
                {
                    'id': 'building_related_data',
                    'required': True,
                    'description': 'A csv file with building related data required to evaluate retrofit strategies',
                    'type': ['incore:multiobjectiveBuildingRelatedData']
                },
                {
                    'id': 'strategy_costs_data',
                    'required': True,
                    'description': 'A csv file with strategy cost data'
                                   'per building',
                    'type': ['incore:multiobjectiveStrategyCosts']
                },
            ],
            'output_datasets': [
                {
                    'id': 'optimal_solution_dv_x',
                    'parent_type': '',
                    'description': 'Optimal solution for decision variable x',
                    'type': 'incore:multiobjectiveOptimalSolutionX'
                },
                {
                    'id': 'optimal_solution_dv_y',
                    'parent_type': '',
                    'description': 'Optimal solution for decision variable y with initial and final retrofitted '
                                   'strategies',
                    'type': 'incore:multiobjectiveOptimalSolutionY'
                }
            ]
        }



from pyincore import IncoreClient
from pyincore.analyses.transportationrecovery import TransportationRecovery


def run_with_base_class():
    client = IncoreClient()
    transportation_recovery = TransportationRecovery(client)

    nodes = "5c5de1dec5c0e488fc0355f7"
    transportation_recovery.load_remote_input_dataset("nodes", nodes)

    links = "5c5de25ec5c0e488fc035613"
    transportation_recovery.load_remote_input_dataset("links", links)

    bridges = "5a284f2dc7d30d13bc082040"
    transportation_recovery.load_remote_input_dataset('bridges', bridges)

    bridge_damage = "5c5ddff0c5c0e488fc0355df"
    transportation_recovery.load_remote_input_dataset('bridge_damage_value', bridge_damage)

    unrepaired = "5c5de0c5c5c0e488fc0355eb"
    transportation_recovery.load_remote_input_dataset('unrepaired_bridge', unrepaired)

    ADT_data = "5c5dde00c5c0e488fc032d7f"
    transportation_recovery.load_remote_input_dataset('ADT', ADT_data)

    transportation_recovery.set_parameter("num_cpu", 4)
    transportation_recovery.set_parameter("pm", 1)
    transportation_recovery.set_parameter('ini_num_population', 5)
    transportation_recovery.set_parameter("population_size", 3)
    transportation_recovery.set_parameter("num_generation", 2)
    transportation_recovery.set_parameter("mutation_rate", 0.1)
    transportation_recovery.set_parameter("crossover_rate", 1.0)

    transportation_recovery.run_analysis()


if __name__ == '__main__':
    run_with_base_class()



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import pandas as pd
import networkx as nx


class EpnFunctionalityUtil:

    @staticmethod
    def get_bad_edges(G, nodestate, linkstate=None, scol='s0'):
        badnodes = nodestate.loc[nodestate.loc[:, scol] == 0, 'nodenwid'].values
        if linkstate is not None:
            badlinks = linkstate.loc[linkstate.loc[:, scol] == 0, ['fromnode', 'tonode']].values
            badlinks = list(zip(badlinks[:, 0], badlinks[:, 1]))
        else:
            badlinks = []
        badlinks2 = list(G.edges(badnodes))
        badlinks.extend(badlinks2)
        return list(set(badlinks))

    @staticmethod
    def network_shortest_paths(G, sources, sinks, weightcol='weight'):
        return pd.Series(nx.multi_source_dijkstra_path_length(G, sources, cutoff=None, weight=weightcol))[sinks]



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import json
from typing import List
from urllib.parse import urljoin
import numpy

import pyincore.globals as pyglobals
from pyincore.utils import return_http_response
from pyincore import IncoreClient

logger = pyglobals.LOGGER


class HazardService:
    """Hazard service client

    Args:
        client (IncoreClient): Service authentication.

    """

    def __init__(self, client: IncoreClient):
        self.client = client

        self.base_earthquake_url = urljoin(client.service_url,
                                           'hazard/api/earthquakes/')
        self.base_tornado_url = urljoin(client.service_url,
                                        'hazard/api/tornadoes/')
        self.base_tsunami_url = urljoin(client.service_url,
                                        'hazard/api/tsunamis/')
        self.base_hurricane_url = urljoin(client.service_url, 'hazard/api/hurricanes/')
        self.base_hurricanewf_url = urljoin(client.service_url,
                                            'hazard/api/hurricaneWindfields/')
        self.base_flood_url = urljoin(client.service_url, 'hazard/api/floods/')

    def get_earthquake_hazard_metadata_list(self, skip: int = None, limit: int = None, space: str = None,
                                            timeout: tuple = (30, 600), **kwargs):
        """Retrieve earthquake metadata list from hazard service. Hazard API endpoint is called.

        Args:
            skip (int):  Skip the first n results, passed to the parameter "skip". Dafault None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit". Dafault None.
            space (str): User's namespace on the server, passed to the parameter "space". Dafault None.
            timeout (tuple): Timeout for the request in seconds. Default (30, 600).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            json: Response containing the metadata.

        """
        url = self.base_earthquake_url
        payload = {}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit
        if space is not None:
            payload['space'] = space

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_earthquake_hazard_metadata(self, hazard_id: str, timeout=(30, 600), **kwargs):
        """Retrieve earthquake metadata from hazard service. Hazard API endpoint is called.

        Args:
            hazard_id (str): ID of the Earthquake.
            timeout (tuple): Timeout for the request in seconds. Default (30, 600).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            json: Response containing the metadata.

        """
        url = urljoin(self.base_earthquake_url, hazard_id)
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_earthquake_hazard_value_set(self, hazard_id: str, demand_type: str,
                                        demand_unit: str, bbox,
                                        grid_spacing: float,
                                        timeout=(30, 600), **kwargs):
        """Retrieve earthquake hazard value set from the Hazard service.

        Args:
            hazard_id (str): ID of the Earthquake.
            demand_type (str):  Ground motion demand type. Examples: PGA, PGV, 0.2 SA
            demand_unit (str): Ground motion demand unit. Examples: g, %g, cm/s, etc.
            bbox (list): Bounding box, list of points.
            grid_spacing (float): Grid spacing.
            timeout (tuple): Timeout for the request in seconds. Default (30, 600).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            np.array: X values..
            np.array: Y values.
            np.array: Hazard value.

        """
        # bbox: [[minx, miny],[maxx, maxy]]
        # raster?demandType=0.2+SA&demandUnits=g&minX=-90.3099&minY=34.9942&maxX=-89.6231&maxY=35.4129&gridSpacing=0.01696
        # bbox
        url = urljoin(self.base_earthquake_url,
                      hazard_id + "/raster")
        payload = {'demandType': demand_type, 'demandUnits': demand_unit,
                   'minX': bbox[0][0], 'minY': bbox[0][1],
                   'maxX': bbox[1][0], 'maxY': bbox[1][1],
                   'gridSpacing': grid_spacing}
        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)
        response = return_http_response(r).json()

        # TODO: need to handle error with the request
        xlist = []
        ylist = []
        zlist = []
        for entry in response['hazardResults']:
            xlist.append(float(entry['longitude']))
            ylist.append(float(entry['latitude']))
            zlist.append(float(entry['hazardValue']))
        x = numpy.array(xlist)
        y = numpy.array(ylist)
        hazard_val = numpy.array(zlist)

        return x, y, hazard_val

    def post_earthquake_hazard_values(self, hazard_id: str, payload: list, amplify_hazard=True, timeout=(30, 600),
                                      **kwargs):
        """ Retrieve bulk hurricane hazard values from the Hazard service.

        Args:
            hazard_id (str): ID of the Earthquake.
            payload (list):
            [
                {
                    "demands": ["PGA", "1.0 SD", "0.9 SA", "1.0 SA", "PGV"],
                    "units": ["g", "cm", "g", "g", "in/s"],
                    "loc": "35.84,-89.90"
                },
                {
                    "demands": ["1.0 SD", "1.0 SA"],
                    "units": ["cm", "g"],
                    "amplifyHazards": [false, false],
                    "loc": "35.84,-89.90"
                }
            ]
        Returns:
            obj: Hazard values.

        """
        url = urljoin(self.base_earthquake_url, hazard_id + "/values")
        kwargs = {"files": {('points', json.dumps(payload)), ('amplifyHazard', json.dumps(amplify_hazard))}}
        r = self.client.post(url, **kwargs)

        return return_http_response(r).json()

    def get_liquefaction_values(self, hazard_id: str, geology_dataset_id: str,
                                demand_unit: str, points: List, timeout=(30, 600), **kwargs):
        """Retrieve earthquake liquefaction values.

        Args:
            hazard_id (str): ID of the Earthquake.
            geology_dataset_id (str):  ID of the geology dataset.
            demand_unit (str): Ground motion demand unit. Examples: g, %g, cm/s, etc.
            points (list): List of points provided as lat,long.
            timeout (tuple): Timeout for the request in seconds. Default (30, 600).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            obj: HTTP response.

        """
        url = urljoin(self.base_earthquake_url,
                      hazard_id + "/liquefaction/values")
        payload = {'demandUnits': demand_unit,
                   'geologyDataset': geology_dataset_id, 'point': points}
        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)
        response = r.json()
        return response

    def post_liquefaction_values(self, hazard_id: str, geology_dataset_id: str, payload: list, timeout=(30, 600),
                                 **kwargs):
        """ Retrieve bulk earthquake liquefaction hazard values from the Hazard service.

        Args:
            hazard_id (str): ID of the Tornado.
            payload (list):
            timeout (tuple): Timeout for the request in seconds. Default (30, 600).
            **kwargs: Arbitrary keyword arguments.
        Returns:
            obj: Hazard values.

        """
        url = urljoin(self.base_earthquake_url, hazard_id + "/liquefaction/values")
        kwargs = {"files": {('points', json.dumps(payload)), ('geologyDataset', geology_dataset_id)}}
        r = self.client.post(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_soil_amplification_value(self, method: str, dataset_id: str,
                                     site_lat: float, site_long: float,
                                     demand_type: str, hazard: float,
                                     default_site_class: str,
                                     timeout=(30, 600), **kwargs):
        """Retrieve earthquake liquefaction values.

        Args:
            method (str): ID of the Earthquake.
            dataset_id (str): ID of the Dataset.
            site_lat (float): Latitude of a point.
            site_long (float): Longitude of a point.
            demand_type (str):  Ground motion demand type. Examples: PGA, PGV, 0.2 SA
            hazard (float): Hazard value.
            default_site_class (str): Default site classification. Expected A, B, C, D, E or F.
            timeout (tuple): Timeout for the request in seconds. Default (30, 600).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            obj: HTTP response.

        """
        url = urljoin(self.base_earthquake_url,
                      'soil/amplification')
        payload = {"method": method, "datasetId": dataset_id,
                   "siteLat": site_lat, "siteLong": site_long,
                   "demandType": demand_type, "hazard": hazard,
                   "defaultSiteClass": default_site_class}
        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    # TODO get_slope_amplification_value needed to be implemented on the server side
    # def get_slope_amplification_value(self)

    def get_supported_earthquake_models(self, timeout=(30, 600), **kwargs):
        """Retrieve suported earthquake models.

        Args:
            timeout (tuple): Timeout for the request in seconds. Default (30, 600).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            obj: HTTP response.

        """
        url = urljoin(self.base_earthquake_url, 'models')
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def create_earthquake(self, eq_json, file_paths: List = [], timeout=(30, 600), **kwargs):
        """Create earthquake on the server. POST API endpoint is called.

        Args:
            eq_json (json): Json representing the Earthquake.
            file_paths (list): List of strings pointing to the paths of the datasets. Not needed for
                model based earthquakes. Default empty list.
            timeout (tuple): Timeout for the request in seconds. Default (30, 600).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            obj: HTTP POST Response. Json of the earthquake posted to the server.

        """
        url = self.base_earthquake_url
        eq_data = {('earthquake', eq_json)}

        for file_path in file_paths:
            eq_data.add(('file', open(file_path, 'rb')))
        kwargs = {"files": eq_data}
        r = self.client.post(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def delete_earthquake(self, hazard_id: str, timeout=(30, 600), **kwargs):
        """Delete an earthquake by it's id, and it's associated datasets

        Args:
            hazard_id (str): ID of the Earthquake.
            timeout (tuple): Timeout for the request in seconds. Default (30, 600).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            obj: Json of deleted hazard

        """
        url = urljoin(self.base_earthquake_url, hazard_id)
        r = self.client.delete(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def search_earthquakes(self, text: str, skip: int = None, limit: int = None, timeout=(30, 600), **kwargs):
        """Search earthquakes.

        Args:
            text (str): Text to search by, passed to the parameter "text".
            skip (int):  Skip the first n results, passed to the parameter "skip". Dafault None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit". Dafault None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_earthquake_url, "search")
        payload = {"text": text}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_earthquake_aleatory_uncertainty(self, hazard_id: str, demand_type: str, timeout=(30, 600), **kwargs):
        """ Gets aleatory uncertainty for an earthquake

        Args:
            hazard_id (str): ID of the Earthquake
            demand_type (str):  Ground motion demand type. Examples: PGA, PGV, 0.2 SA
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
             obj: HTTP POST Response with aleatory uncertainty value.

        """
        url = urljoin(self.base_earthquake_url, hazard_id + "/aleatoryuncertainty")
        payload = {"demandType": demand_type}

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def get_earthquake_variance(self, hazard_id: str, variance_type: str, demand_type: str,
                                demand_unit: str, points: List, timeout=(30, 600), **kwargs):
        """Gets total and epistemic variance for a model based earthquake

        Args:
            hazard_id (str): ID of the Earthquake
            variance_type (str): Type of Variance. epistemic or total
            demand_type (str):  Ground motion demand type. Examples: PGA, PGV, 0.2 SA
            demand_unit (str): Demand unit. Examples: g, in
            points (list): List of points provided as lat,long.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP POST Response with variance value.

        """
        url = urljoin(self.base_earthquake_url, hazard_id + "/variance/" + variance_type)
        payload = {"demandType": demand_type, "demandUnits": demand_unit, 'point': points}

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def get_tornado_hazard_metadata_list(self, skip: int = None, limit: int = None, space: str = None,
                                         timeout=(30, 600), **kwargs):
        """Retrieve tornado metadata list from hazard service. Hazard API endpoint is called.

        Args:
            skip (int):  Skip the first n results, passed to the parameter "skip". Dafault None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit". Dafault None.
            space (str): User's namespace on the server, passed to the parameter "space". Dafault None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = self.base_tornado_url
        payload = {}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit
        if space is not None:
            payload['space'] = space

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_tornado_hazard_metadata(self, hazard_id: str, timeout=(30, 600), **kwargs):
        """Retrieve tornado metadata list from hazard service. Hazard API endpoint is called.

        Args:
            hazard_id (str): ID of the Tornado.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = urljoin(self.base_tornado_url, hazard_id)
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def post_tornado_hazard_values(self, hazard_id: str, payload: list, seed=None, timeout=(30, 600), **kwargs):
        """ Retrieve bulk tornado hazard values from the Hazard service.

        Args:
            hazard_id (str): ID of the Tornado.
            payload (list):
            seed: (None or int): Seed value for random values.
        Returns:
            obj: Hazard values.

        """
        url = urljoin(self.base_tornado_url, hazard_id + "/values")

        if seed is not None:
            kwargs["files"] = {('points', json.dumps(payload)), ('seed', json.dumps(seed))}
        else:
            kwargs["files"] = {('points', json.dumps(payload))}

        r = self.client.post(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def create_tornado_scenario(self, tornado_json, file_paths: List = [], timeout=(30, 600), **kwargs):
        """Create tornado on the server. POST API endpoint is called.

        Args:
            tornado_json (json): JSON representing the tornado.
            file_paths (list): List of strings pointing to the paths of the shape files. Not needed for
                model based tornadoes.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP POST Response. Json of the created tornado.

        """
        url = self.base_tornado_url
        tornado_data = {('tornado', tornado_json)}

        for file_path in file_paths:
            tornado_data.add(('file', open(file_path, 'rb')))
        kwargs = {"files": tornado_data}
        r = self.client.post(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def delete_tornado(self, hazard_id: str, timeout=(30, 600), **kwargs):
        """Delete a tornado by it's id, and it's associated datasets

        Args:
            hazard_id (str): ID of the Tornado.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: Json of deleted hazard

        """
        url = urljoin(self.base_tornado_url, hazard_id)
        r = self.client.delete(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def search_tornadoes(self, text: str, skip: int = None, limit: int = None, timeout=(30, 600), **kwargs):
        """Search tornadoes.

        Args:
            text (str): Text to search by, passed to the parameter "text".
            skip (int):  Skip the first n results, passed to the parameter "skip". Dafault None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit". Dafault None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_tornado_url, "search")
        payload = {"text": text}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_tsunami_hazard_metadata_list(self, skip: int = None, limit: int = None, space: str = None,
                                         timeout=(30, 600), **kwargs):
        """Retrieve tsunami metadata list from hazard service. Hazard API endpoint is called.

        Args:
            skip (int): Skip the first n results, passed to the parameter "skip". Dafault None.
            limit (int): Limit number of results to return. Passed to the parameter "limit". Dafault None.
            space (str): User's namespace on the server, passed to the parameter "space". Dafault None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = self.base_tsunami_url
        payload = {}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit
        if space is not None:
            payload['space'] = space

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_tsunami_hazard_metadata(self, hazard_id: str, timeout=(30, 600), **kwargs):
        """Retrieve tsunami metadata list from hazard service. Hazard API endpoint is called.

        Args:
            hazard_id (str): ID of the Tsunami.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = urljoin(self.base_tsunami_url, hazard_id)
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def post_tsunami_hazard_values(self, hazard_id: str, payload: list, timeout=(30, 600), **kwargs):
        """ Retrieve bulk tsunami hazard values from the Hazard service.

        Args:
            hazard_id (str): ID of the Tsunami.
            payload (list):
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.
        Returns:
            obj: Hazard values.

        """
        url = urljoin(self.base_tsunami_url, hazard_id + "/values")
        kwargs = {"files": {('points', json.dumps(payload))}}
        r = self.client.post(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def create_tsunami_hazard(self, tsunami_json, file_paths: List, timeout=(30, 600), **kwargs):
        """Create tsunami on the server. POST API endpoint is called.

        Args:
            tsunami_json: JSON representing the tsunami.
            file_paths: List of strings pointing to the paths of the datasets.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP POST Response. Json of the created tsunami.

        """

        url = self.base_tsunami_url
        tsunami_data = {('tsunami', tsunami_json)}

        for file_path in file_paths:
            tsunami_data.add(('file', open(file_path, 'rb')))
        kwargs = {"files": tsunami_data}
        r = self.client.post(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def delete_tsunami(self, hazard_id: str, timeout=(30, 600), **kwargs):
        """Delete a tsunami by it's id, and it's associated datasets

        Args:
            hazard_id (str): ID of the Tsunami.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: Json of deleted hazard

        """
        url = urljoin(self.base_tsunami_url, hazard_id)
        r = self.client.delete(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def search_tsunamis(self, text: str, skip: int = None, limit: int = None, timeout=(30, 600), **kwargs):
        """Search tsunamis.

        Args:
            text (str): Text to search by, passed to the parameter "text".
            skip (int):  Skip the first n results, passed to the parameter "skip". Default None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit". Dafault None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_tsunami_url, "search")
        payload = {"text": text}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def create_hurricane(self, hurricane_json, file_paths: List, timeout=(30, 600), **kwargs):
        """Create hurricanes on the server. POST API endpoint is called.

        Args:
            hurricane_json (obj): JSON representing the hurricane.
            file_paths (list): List of strings pointing to the paths of the datasets.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP POST Response. Json of the created hurricane.

        """
        url = self.base_hurricane_url
        hurricane_data = {('hurricane', hurricane_json)}

        for file_path in file_paths:
            hurricane_data.add(('file', open(file_path, 'rb')))
        kwargs = {"files": hurricane_data}
        r = self.client.post(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_hurricane_metadata_list(self, skip: int = None, limit: int = None, space: str = None,
                                    timeout=(30, 600), **kwargs):
        """Retrieve hurricane metadata list from hazard service. Hazard API endpoint is called.

        Args:
            skip (int):  Skip the first n results, passed to the parameter "skip". Default None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit". Default None.
            space (str): User's namespace on the server, passed to the parameter "space". Default None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = self.base_hurricane_url
        payload = {}

        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit
        if space is not None:
            payload['space'] = space

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_hurricane_metadata(self, hazard_id, timeout=(30, 600), **kwargs):
        """Retrieve hurricane metadata list from hazard service. Hazard API endpoint is called.

        Args:
            hazard_id (str): ID of the Hurricane.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = urljoin(self.base_hurricane_url, hazard_id)
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def post_hurricane_hazard_values(self, hazard_id: str, payload: list, timeout=(30, 600), **kwargs):
        """ Retrieve bulk hurricane hazard values from the Hazard service.

        Args:
            hazard_id (str): ID of the Hurricane.
            payload (list):
        Returns:
            obj: Hazard values.

        """
        url = urljoin(self.base_hurricane_url, hazard_id + "/values")
        kwargs = {"files": {('points', json.dumps(payload))}}
        r = self.client.post(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def delete_hurricane(self, hazard_id: str, timeout=(30, 600), **kwargs):
        """Delete a hurricane by it's id, and it's associated datasets

        Args:
            hazard_id (str): ID of the Hurricane.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: Json of deleted hazard

        """
        url = urljoin(self.base_hurricane_url, hazard_id)
        r = self.client.delete(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def search_hurricanes(self, text: str, skip: int = None, limit: int = None, timeout=(30, 600), **kwargs):
        """Search hurricanes.

        Args:
            text (str): Text to search by, passed to the parameter "text".
            skip (int):  Skip the first n results, passed to the parameter "skip", default None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit", default None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.
        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_hurricane_url, "search")
        payload = {"text": text}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def create_flood(self, flood_json, file_paths: List, timeout=(30, 600), **kwargs):
        """Create floods on the server. POST API endpoint is called.

        Args:
            flood_json (obj): JSON representing the flood.

        Returns:
            response (obj): HTTP POST Response. Json of the created flood.

        """
        url = self.base_flood_url
        flood_data = {('flood', flood_json)}

        for file_path in file_paths:
            flood_data.add(('file', open(file_path, 'rb')))
        kwargs = {"files": flood_data}
        r = self.client.post(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_flood_metadata_list(self, skip: int = None, limit: int = None, space: str = None, timeout=(30, 600),
                                **kwargs):
        """Retrieve flood metadata list from hazard service. Hazard API endpoint is called.

        Args:
            skip (int):  Skip the first n results, passed to the parameter "skip". Default None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit". Default None.
            space (str): User's namespace on the server, passed to the parameter "space". Default None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = self.base_flood_url
        payload = {}

        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit
        if space is not None:
            payload['space'] = space

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_flood_metadata(self, hazard_id, timeout=(30, 600), **kwargs):
        """Retrieve flood metadata list from hazard service. Hazard API endpoint is called.

        Args:
            hazard_id (str): ID of the flood.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.
        Returns:
            obj: HTTP response containing the metadata.

        """
        url = urljoin(self.base_flood_url, hazard_id)
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def post_flood_hazard_values(self, hazard_id: str, payload: list, timeout=(30, 600), **kwargs):
        """ Retrieve bulk flood hazard values from the Hazard service.

        Args:
            hazard_id (str): ID of the Flood.
            payload (list):
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.
        Returns:
            obj: Hazard values.

        """
        url = urljoin(self.base_flood_url, hazard_id + "/values")
        kwargs = {"files": {('points', json.dumps(payload))}}
        r = self.client.post(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def delete_flood(self, hazard_id: str, timeout=(30, 600), **kwargs):
        """Delete a flood by it's id, and it's associated datasets

        Args:
            hazard_id (str): ID of the Flood.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: Json of deleted hazard

        """
        url = urljoin(self.base_flood_url, hazard_id)
        r = self.client.delete(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def search_floods(self, text: str, skip: int = None, limit: int = None, timeout=(30, 600), **kwargs):
        """Search floods.

        Args:
            text (str): Text to search by, passed to the parameter "text".
            skip (int):  Skip the first n results, passed to the parameter "skip", default None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit", default None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_flood_url, "search")
        payload = {"text": text}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def create_hurricane_windfield(self, hurr_wf_inputs, timeout=(30, 10800), **kwargs):
        """Create wind fields on the server. POST API endpoint is called.

        Args:
            hurr_wf_inputs (obj): JSON representing the hurricane.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 10800).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP POST Response. Json of the created hurricane.

        """
        url = self.base_hurricanewf_url
        headers = {'Content-type': 'application/json'}
        new_headers = {**self.client.session.headers, **headers}
        kwargs = {"headers": new_headers}
        r = self.client.post(url, data=hurr_wf_inputs, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_hurricanewf_metadata_list(self, coast: str = None, category: int = None, skip: int = None,
                                      limit: int = None, space: str = None, timeout=(30, 600), **kwargs):
        """Retrieve hurricane metadata list from hazard service. Hazard API endpoint is called.

        Args:
            coast (str): Coast, passed to the parameter "coast".  Examples: gulf, florida or east.
            category (int): Hurricane category, passed to the parameter "coast".  Examples: between 1 and 5.
            skip (int):  Skip the first n results, passed to the parameter "skip". Default None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit". Dafault None.
            space (str): User's namespace on the server, passed to the parameter "space". Dafault None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = self.base_hurricanewf_url
        payload = {}

        if coast is not None:
            payload['coast'] = coast
        if category is not None:
            payload['category'] = category
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit
        if space is not None:
            payload['space'] = space

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_hurricanewf_metadata(self, hazard_id, timeout=(30, 600), **kwargs):
        """Retrieve hurricane metadata list from hazard service. Hazard API endpoint is called.

        Args:
            hazard_id (str): ID of the Hurricane.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = urljoin(self.base_hurricanewf_url, hazard_id)
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def post_hurricanewf_hazard_values(self, hazard_id: str, payload: list, elevation: int, roughness: float,
                                       timeout=(30, 600), **kwargs):

        """ Retrieve bulk hurricane windfield hazard values from the Hazard service.

        Args:
            hazard_id (str): ID of the hurricanewf.
            payload (list):
            elevation (int):
            roughness (float):
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.
        Returns:
            obj: Hazard values.

        """
        url = urljoin(self.base_hurricanewf_url, hazard_id + "/values")
        kwargs["files"] = {('points', json.dumps(payload)),
                           ('elevation', json.dumps(elevation)),
                           ('roughness', json.dumps(roughness))}
        r = self.client.post(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_hurricanewf_json(self, coast: str, category: int, trans_d: float, land_fall_loc: int, demand_type: str,
                             demand_unit: str, resolution: int = 6, grid_points: int = 80,
                             rf_method: str = "circular", timeout=(30, 600), **kwargs):
        """Retrieve hurricane wind field values from the Hazard service.

        Args:
            coast (str): Coast, passed to the parameter "coast".  Examples: gulf, florida or east.
            category (int): Hurricane category, passed to the parameter "coast".  Examples: between 1 and 5.
            trans_d (float): Trans_d.
            land_fall_loc (float):
            demand_type (str):
            demand_unit (str):
            resolution (int): Resolution, default 6.
            grid_points (int): Grid points, default 80.
            rf_method (str): Rf method, Default "circular"
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response.

        """
        # land_fall_loc: IncorePoint e.g.'28.01, -83.85'
        url = urljoin(self.base_hurricanewf_url, "json/" + coast)
        payload = {"category": category, "TransD": trans_d,
                   "LandfallLoc": land_fall_loc,
                   "demandType": demand_type, "demandUnits": demand_unit,
                   "resolution": resolution, "gridPoints": grid_points,
                   "reductionType": rf_method}
        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def delete_hurricanewf(self, hazard_id: str, timeout=(30, 600), **kwargs):
        """Delete a hurricane windfield by it's id, and it's associated datasets

        Args:
            hazard_id (str): ID of the Hurricane Windfield.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: Json of deleted hazard

        """
        url = urljoin(self.base_hurricanewf_url, hazard_id)
        r = self.client.delete(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def search_hurricanewf(self, text: str, skip: int = None, limit: int = None, timeout=(30, 600), **kwargs):
        """Search hurricanes.

        Args:
            text (str): Text to search by, passed to the parameter "text".
            skip (int):  Skip the first n results, passed to the parameter "skip", default None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit", default None.
            timeout (tuple):  Timeout for the request. Passed to the parameter "timeout". Dafault (30, 600).
            **kwargs:  Additional keyword arguments.

        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_hurricanewf_url, "search")
        payload = {"text": text}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    # TODO replace this with API endpoint in the future
    def get_allowed_demands(self, hazard_type, timeout=(30, 600), **kwargs):
        if self.client.offline:
            if hazard_type in HazardConstant.DEFAULT_ALLOWED_DEMANDS.keys():
                return HazardConstant.DEFAULT_ALLOWED_DEMANDS.get(hazard_type)
            else:
                raise ValueError("Unknown hazard type!")
        else:
            if hazard_type == 'earthquake':
                url = urljoin(self.base_earthquake_url, "demands")
            elif hazard_type == 'tornado':
                url = urljoin(self.base_tornado_url, "demands")
            elif hazard_type == 'tsunami':
                url = urljoin(self.base_tsunami_url, "demands")
            elif hazard_type == 'hurricane':
                url = urljoin(self.base_hurricane_url, "demands")
            elif hazard_type == 'hurricaneWindfield':
                url = urljoin(self.base_hurricanewf_url, "demands")
            elif hazard_type == 'flood':
                url = urljoin(self.base_flood_url, "demands")
            else:
                raise ValueError("Unknown hazard type!")

            r = self.client.get(url, timeout=timeout, **kwargs)
            return return_http_response(r).json()


class HazardConstant:

    """HazardConstant class to hold all the constants related to hazard."""
    DEFAULT_ALLOWED_DEMANDS = {
        "earthquake": [
            {
                "demand_type": "pga",
                "demand_unit": [
                    "g",
                    "in/sec^2",
                    "m/sec^2"
                ],
                "description": "Peak ground acceleration"
            },
            {
                "demand_type": "pgv",
                "demand_unit": [
                    "in/s",
                    "cm/s"
                ],
                "description": "Peak ground velocity"
            },
            {
                "demand_type": "pgd",
                "demand_unit": [
                    "in",
                    "ft",
                    "m"
                ],
                "description": "Peak ground displacement"
            },
            {
                "demand_type": "sa",
                "demand_unit": [
                    "g",
                    "in/sec^2",
                    "m/sec^2"
                ],
                "description": "Spectral acceleration"
            },
            {
                "demand_type": "sd",
                "demand_unit": [
                    "in",
                    "ft",
                    "m",
                    "cm"
                ],
                "description": "Spectral displacement"
            },
            {
                "demand_type": "sv",
                "demand_unit": [
                    "cm/s",
                    "in/s"
                ],
                "description": "Spectral Velocity"
            }
        ],
        "tsunami": [
            {
                "demand_type": "Hmax",
                "demand_unit": [
                    "ft",
                    "m"
                ],
                "description": "Onshore: maximum tsunami height above local ground level overland. Offshore: "
                               "maximum tsunami height taken crest to trough"
            },
            {
                "demand_type": "Vmax",
                "demand_unit": [
                    "mph",
                    "kph",
                    "ft/sec",
                    "m/sec"
                ],
                "description": "Maximum near-coast or overland water velocity due to tsunami"
            },
            {
                "demand_type": "Mmax",
                "demand_unit": [
                    "m^3/s^2",
                    "ft^3/s^2"
                ],
                "description": ""
            }
        ],
        "flood": [
            {
                "demand_type": "inundationDepth",
                "demand_unit": [
                    "ft",
                    "m"
                ],
                "description": "Depth of the water surface relative to local ground level"
            },
            {
                "demand_type": "waterSurfaceElevation",
                "demand_unit": [
                    "ft",
                    "m"
                ],
                "description": "Elevation of the water surface above reference datum (e.g. NAVD88, mean sea level)"
            }
        ],
        "tornado": [
            {
                "demand_type": "wind",
                "demand_unit": [
                    "mps",
                    "mph"
                ],
                "description": "Defined as a wind velocity below"
            }
        ],
        "hurricaneWindfield": [
            {
                "demand_type": "3s",
                "demand_unit": [
                    "kph",
                    "mph",
                    "kt"
                ],
                "description": "Typically, reported at 10 m above local ground or sea level"
            },
            {
                "demand_type": "60s",
                "demand_unit": [
                    "kph",
                    "mph",
                    "kt"
                ],
                "description": "Typically, reported at 10 m above local ground or sea level"
            }
        ],
        "hurricane": [
            {
                "demand_type": "waveHeight",
                "demand_unit": [
                    "ft",
                    "m",
                    "in",
                    "cm"
                ],
                "description": " Height of wave measured crest to trough.  Characteristic wave height is typically the  "
                               "average of one third highest waves for a random sea."
            },
            {
                "demand_type": "surgeLevel",
                "demand_unit": [
                    "ft",
                    "m",
                    "in",
                    "cm"
                ],
                "description": "Elevation of the water surface above reference datum (e.g. NAVD88, mean sea level)"
            },
            {
                "demand_type": "inundationDuration",
                "demand_unit": [
                    "hr",
                    "min",
                    "s"
                ],
                "description": "Time that inundation depth exceeds a critical threshold for a given storm"
            },
            {
                "demand_type": "inundationDepth",
                "demand_unit": [
                    "ft",
                    "m",
                    "in",
                    "cm"
                ],
                "description": "Depth of the water surface relative to local ground level"
            },
            {
                "demand_type": "wavePeriod",
                "demand_unit": [
                    "s",
                    "hr",
                    "min"
                ],
                "description": "Time between wave crests.  Characteristic wave period is typically the inverse of the "
                               "spectral peak frequency for a random sea"
            },
            {
                "demand_type": "waveDirection",
                "demand_unit": [
                    "deg",
                    "rad"
                ],
                "description": "Principle wave direction associated with the characteristic wave height and period"
            },
            {
                "demand_type": "waterVelocity",
                "demand_unit": [
                    "ft/s",
                    "m/s",
                    "in/s"
                ],
                "description": ""
            },
            {
                "demand_type": "windVelocity",
                "demand_unit": [
                    "ft/s",
                    "m/s",
                    "m/sec",
                    "in/s"
                ],
                "description": ""
            }
        ],
        "earthquake+tsunami": [
            {
                "demand_type": "pga",
                "demand_unit": [
                    "g",
                    "in/sec^2",
                    "m/sec^2"
                ],
                "description": "Peak ground acceleration"
            },
            {
                "demand_type": "pgv",
                "demand_unit": [
                    "in/s",
                    "cm/s"
                ],
                "description": "Peak ground velocity"
            },
            {
                "demand_type": "pgd",
                "demand_unit": [
                    "in",
                    "ft",
                    "m"
                ],
                "description": "Peak ground displacement"
            },
            {
                "demand_type": "sa",
                "demand_unit": [
                    "g",
                    "in/sec^2",
                    "m/sec^2"
                ],
                "description": "Spectral acceleration"
            },
            {
                "demand_type": "sd",
                "demand_unit": [
                    "in",
                    "ft",
                    "m",
                    "cm"
                ],
                "description": "Spectral displacement"
            },
            {
                "demand_type": "sv",
                "demand_unit": [
                    "cm/s",
                    "in/s"
                ],
                "description": "Spectral Velocity"
            },
            {
                "demand_type": "Hmax",
                "demand_unit": [
                    "ft",
                    "m"
                ],
                "description": "Onshore: maximum tsunami height above local ground level overland. Offshore: maximum tsunami height taken crest to trough"
            },
            {
                "demand_type": "Vmax",
                "demand_unit": [
                    "mph",
                    "kph",
                    "ft/sec",
                    "m/sec"
                ],
                "description": "Maximum near-coast or overland water velocity due to tsunami"
            },
            {
                "demand_type": "Mmax",
                "demand_unit": [
                    "m^3/s^2",
                    "ft^3/s^2"
                ],
                "description": ""
            }
        ]
    }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import pandas as pd
import traceback

from pyincore import BaseAnalysis


class BuildingEconLoss(BaseAnalysis):
    """Direct Building Economic Loss analysis calculates the building loss based on
    building appraisal value, mean damage and an inflation multiplier from user's input.
    We are not implementing any inflation calculation based on consumer price indices
    at the moment. A user must supply the inflation percentage between a building appraisal
    year and year of interest (current, date of hazard etc.)

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        # percentage
        self.infl_factor = 0.0

        super(BuildingEconLoss, self).__init__(incore_client)

    def run(self):
        """Executes building economic damage analysis."""

        # Get inflation input in %
        self.infl_factor = self.get_parameter("inflation_factor")
        if self.infl_factor is None:
            self.infl_factor = 0.0

        # Building dataset
        bldg_set = self.get_input_dataset("buildings").get_inventory_reader()

        # Occupancy type of the exposure
        occ_multiplier = self.get_input_dataset("occupancy_multiplier")
        if occ_multiplier is not None:
            occ_multiplier = occ_multiplier.get_csv_reader()
            occ_mult_df = pd.DataFrame(occ_multiplier)
        else:
            occ_mult_df = None

        try:
            prop_select = []
            for bldg_item in list(bldg_set):
                guid = bldg_item["properties"]["guid"]
                appr_bldg = bldg_item["properties"]["appr_bldg"]
                year_built = bldg_item["properties"]["year_built"]
                occ_type = bldg_item["properties"]["occ_type"]
                prop_select.append([guid, year_built, occ_type, appr_bldg])

            bldg_set_df = pd.DataFrame(prop_select, columns=["guid", "year_built", "occ_type", "appr_bldg"])
            bldg_dmg_set = self.get_input_dataset("building_mean_dmg").get_csv_reader()
            bldg_dmg_df = pd.DataFrame(list(bldg_dmg_set))

            dmg_set_df = pd.merge(bldg_set_df, bldg_dmg_df, how="outer", left_on="guid", right_on="guid",
                                  sort=True, copy=True)
            infl_mult = self.get_inflation_mult()

            dmg_set_df = self.add_multipliers(dmg_set_df, occ_mult_df)

            bldg_results = dmg_set_df[["guid"]].copy()
            loss = 0.0
            lossdev = 0.0

            if "appr_bldg" in dmg_set_df:
                loss = dmg_set_df["appr_bldg"].astype(float) * dmg_set_df["meandamage"].astype(float) * dmg_set_df[
                    "Multiplier"].astype(float) * infl_mult
                lossdev = dmg_set_df["appr_bldg"].astype(float) * dmg_set_df["mdamagedev"].astype(float) * dmg_set_df[
                    "Multiplier"].astype(float) * infl_mult

            bldg_results["loss"] = loss.round(2)
            bldg_results["loss_dev"] = lossdev.round(2)

            result_name = self.get_parameter("result_name")

            self.set_result_csv_data("result", bldg_results, result_name, "dataframe")
            return True
        except Exception as e:
            # This prints the type, value and stacktrace of error being handled.
            traceback.print_exc()
            raise e

    def get_inflation_mult(self):
        """Get inflation multiplier from user's input.

        Returns:
            float: Inflation multiplier.

        """
        return (self.infl_factor / 100.0) + 1.0

    def add_multipliers(self, dmg_set_df, occ_mult_df):
        """Add occupancy multipliers to damage dataset.

        Args:
            dmg_set_df (pd.DataFrame): Building inventory dataset with guid and mean damages.
            occ_mult_df (pd.DataFrame): Occupation multiplier set.

        Returns:
            pd.DataFrame: Merged inventory.

        """
        if occ_mult_df is not None:
            # Occupancy multipliers are in percentages, convert to multiplication factors
            occ_mult_df["Multiplier"] = (occ_mult_df["Multiplier"].astype(float) / 100.0) + 1.0

            occ_mult_df = occ_mult_df.rename(columns={"Occupancy": "occ_type"})
            dmg_set_df = pd.merge(dmg_set_df, occ_mult_df, how="left", left_on="occ_type",
                                  right_on="occ_type", sort=True, copy=True)
        else:
            dmg_set_df["Multiplier"] = 1.0

        return dmg_set_df

    def get_spec(self):
        """Get specifications of the building damage analysis.

        Returns:
            obj: A JSON object of specifications of the building damage analysis.

        """
        return {
            'name': 'building-economy-damage',
            'description': 'building economy damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'inflation_factor',
                    'required': False,
                    'description': 'Inflation factor to adjust the appraisal values of buildings. Default 0.0',
                    'type': float
                },
            ],
            'input_datasets': [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'Building Inventory',
                    'type': ['ergo:buildingInventory', 'ergo:buildingInventoryVer4',
                             'ergo:buildingInventoryVer5', 'ergo:buildingInventoryVer6',
                             'ergo:buildingInventoryVer7']
                },
                {
                    'id': 'building_mean_dmg',
                    'required': True,
                    'description': 'A CSV file with building mean damage results for either Structural, '
                                   'Drift-Sensitive Nonstructural, Acceleration-Sensitive Nonstructural '
                                   'or Contents Damage component.',
                    'type': ['ergo:meanDamage']
                },
                {
                    'id': 'occupancy_multiplier',
                    'required': False,
                    'description': 'Building occupancy damage multipliers. These percentage multipliers account '
                                   'for the value associated with different types of components (structural, '
                                   'acceleration-sensitive nonstructural, '
                                   'drift-sensitive nonstructural, contents).',
                    'type': ['incore:buildingOccupancyMultiplier']
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'buildings',
                    'description': 'CSV file of building economy damages',
                    'type': 'incore:buildingEconomicLoss'
                }
            ]
        }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


"""Buried Pipeline Damage Analysis with Repair Rate Calculation

"""

import concurrent.futures
import math
from itertools import repeat

from pyincore import BaseAnalysis, HazardService, FragilityService, \
    AnalysisUtil, GeoUtil
from pyincore.analyses.pipelinedamagerepairrate.pipelineutil import \
    PipelineUtil


class PipelineDamageRepairRate(BaseAnalysis):
    """Computes pipeline damage for a hazard.

    Args:
        incore_client: Service client with authentication info

    """

    def __init__(self, incore_client):
        self.hazardsvc = HazardService(incore_client)
        self.fragilitysvc = FragilityService(incore_client)

        super(PipelineDamageRepairRate, self).__init__(incore_client)

    def run(self):
        """Execute pipeline damage analysis """
        # Pipeline dataset
        pipeline_dataset = self.get_input_dataset(
            "pipeline").get_inventory_reader()

        # get input hazard
        hazard, hazard_type, hazard_dataset_id = self.create_hazard_object_from_input_params()

        user_defined_cpu = 1

        if not self.get_parameter("num_cpu") is None and self.get_parameter(
                "num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        dataset_size = len(pipeline_dataset)
        num_workers = AnalysisUtil.determine_parallelism_locally(self,
                                                                 dataset_size,
                                                                 user_defined_cpu)

        avg_bulk_input_size = int(dataset_size / num_workers)
        inventory_args = []
        count = 0
        inventory_list = list(pipeline_dataset)
        while count < len(inventory_list):
            inventory_args.append(
                inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        (ds_results, damage_results) = self.pipeline_damage_concurrent_future(
            self.pipeline_damage_analysis_bulk_input, num_workers,
            inventory_args, repeat(hazard), repeat(hazard_type), repeat(hazard_dataset_id))

        self.set_result_csv_data("result", ds_results, name=self.get_parameter("result_name"))
        self.set_result_json_data("metadata",
                                  damage_results,
                                  name=self.get_parameter("result_name") + "_additional_info")

        return True

    def pipeline_damage_concurrent_future(self, function_name, num_workers,
                                          *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            num_workers (int): Maximum number workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            list: A list of ordered dictionaries with building damage values and other data/metadata.

        """
        output_ds = []
        output_dmg = []
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_workers) as executor:
            for ret1, ret2 in executor.map(function_name, *args):
                output_ds.extend(ret1)
                output_dmg.extend(ret2)

        return output_ds, output_dmg

    def pipeline_damage_analysis_bulk_input(self, pipelines, hazard, hazard_type, hazard_dataset_id):
        """Run pipeline damage analysis for multiple pipelines.

        Args:
            pipelines (list): multiple pipelines from pieline dataset.
            hazard (obj): Hazard object
            hazard_type (str): Hazard type
            hazard_dataset_id (str): An id of the hazard exposure.

        Returns:
            ds_results (list): A list of ordered dictionaries with pipeline damage values and other data/metadata.
            damage_results (list): A list of ordered dictionaries with pipeline damage metadata.
        """
        # Get Fragility key
        fragility_key = self.get_parameter("fragility_key")
        if fragility_key is None:
            fragility_key = PipelineUtil.DEFAULT_TSU_FRAGILITY_KEY if hazard_type == 'tsunami' else \
                PipelineUtil.DEFAULT_EQ_FRAGILITY_KEY
            self.set_parameter("fragility_key", fragility_key)

        # get fragility set
        fragility_sets = self.fragilitysvc.match_inventory(
            self.get_input_dataset("dfr3_mapping_set"), pipelines, fragility_key)

        # Get Liquefaction Fragility Key
        liquefaction_fragility_key = self.get_parameter(
            "liquefaction_fragility_key")
        if hazard_type == "earthquake" and liquefaction_fragility_key is None:
            liquefaction_fragility_key = PipelineUtil.LIQ_FRAGILITY_KEY

        # Liquefaction
        use_liquefaction = False
        if hazard_type == "earthquake" and self.get_parameter(
                "use_liquefaction") is not None:
            use_liquefaction = self.get_parameter("use_liquefaction")

        # Get geology dataset id
        geology_dataset_id = self.get_parameter(
            "liquefaction_geology_dataset_id")
        fragility_sets_liq = None
        if geology_dataset_id is not None:
            fragility_sets_liq = self.fragilitysvc.match_inventory(
                self.get_input_dataset("dfr3_mapping_set"), pipelines,
                liquefaction_fragility_key)

        values_payload = []
        values_payload_liq = []  # for liquefaction if used
        unmapped_pipelines = []
        mapped_pipelines = []
        for pipeline in pipelines:
            # if find a match fragility for that pipeline
            if pipeline["id"] in fragility_sets.keys():
                fragility_set = fragility_sets[pipeline["id"]]
                location = GeoUtil.get_location(pipeline)
                loc = str(location.y) + "," + str(location.x)
                demands = fragility_set.demand_types
                units = fragility_set.demand_units
                value = {
                    "demands": demands,
                    "units": units,
                    "loc": loc
                }
                values_payload.append(value)
                mapped_pipelines.append(pipeline)

                # Check if liquefaction is applicable
                if use_liquefaction and \
                        geology_dataset_id is not None and \
                        fragility_sets_liq is not None and \
                        pipeline["id"] in fragility_sets_liq:
                    fragility_set_liq = fragility_sets_liq[pipeline["id"]]
                    demands_liq = fragility_set_liq.demand_types
                    units_liq = fragility_set_liq.demand_units
                    value_liq = {
                        "demands": demands_liq,
                        "units": units_liq,
                        "loc": loc
                    }
                    values_payload_liq.append(value_liq)
            else:
                unmapped_pipelines.append(pipeline)
        del pipelines

        if hazard_type == 'earthquake' or hazard_type == 'tsunami':
            hazard_resp = hazard.read_hazard_values(values_payload, self.hazardsvc)
        else:
            raise ValueError("The provided hazard type is not supported yet by this analysis")

        liquefaction_resp = None
        # Check if liquefaction is applicable
        if use_liquefaction is True and \
                fragility_sets_liq is not None and \
                geology_dataset_id is not None:
            liquefaction_resp = self.hazardsvc.post_liquefaction_values(hazard_dataset_id, geology_dataset_id,
                                                                        values_payload_liq)

        # calculate LS and DS
        ds_results = []
        damage_results = []
        for i, pipeline in enumerate(mapped_pipelines):
            # default
            pgv_repairs = None
            pgd_repairs = 0.0
            total_repair_rate = None
            break_rate = None
            leak_rate = None
            failure_probability = None
            num_pgv_repairs = None
            num_pgd_repairs = 0.0
            num_repairs = None

            liq_hazard_vals = None
            liq_demand_types = None
            liq_demand_units = None
            liquefaction_prob = None

            ds_result = dict()
            damage_result = dict()
            ds_result['guid'] = pipeline['properties']['guid']
            damage_result['guid'] = pipeline['properties']['guid']

            fragility_set = fragility_sets[pipeline["id"]]
            # TODO assume there is only one curve
            fragility_curve = fragility_set.fragility_curves[0]

            hazard_vals = AnalysisUtil.update_precision_of_lists(hazard_resp[i]["hazardValues"])
            demand_types = hazard_resp[i]["demands"]
            demand_units = hazard_resp[i]["units"]

            hval_dict = dict()
            for j, d in enumerate(fragility_set.demand_types):
                hval_dict[d] = hazard_vals[j]

            if not AnalysisUtil.do_hazard_values_have_errors(hazard_resp[i]["hazardValues"]):
                pipeline_args = fragility_set.construct_expression_args_from_inventory(pipeline)
                pgv_repairs = \
                    fragility_curve.solve_curve_expression(
                        hval_dict, fragility_set.curve_parameters, **pipeline_args)
                # Convert PGV repairs to SI units
                pgv_repairs = PipelineUtil.convert_result_unit(fragility_curve.return_type["unit"], pgv_repairs)

                length = PipelineUtil.get_pipe_length(pipeline)

                # Number of PGV repairs
                num_pgv_repairs = pgv_repairs * length

                # Check if liquefaction is applicable
                if use_liquefaction is True \
                        and fragility_sets_liq is not None \
                        and geology_dataset_id is not None \
                        and liquefaction_resp is not None:
                    fragility_set_liq = fragility_sets_liq[pipeline["id"]]

                    # TODO assume there is only one curve
                    liq_fragility_curve = fragility_set_liq.fragility_curves[0]

                    liq_hazard_vals = AnalysisUtil.update_precision_of_lists(liquefaction_resp[i]["pgdValues"])
                    liq_demand_types = liquefaction_resp[i]["demands"]
                    liq_demand_units = liquefaction_resp[i]["units"]
                    liquefaction_prob = liquefaction_resp[i]['liqProbability']
                    liq_hval_dict = dict()
                    for j, d in enumerate(liquefaction_resp[i]["demands"]):
                        liq_hval_dict[d] = liq_hazard_vals[j]

                    # !important! removing the liqProbability and passing in the "diameter"
                    # no fragility is actually using liqProbability
                    pipeline_args = fragility_set_liq.construct_expression_args_from_inventory(pipeline)
                    pgd_repairs = \
                        liq_fragility_curve.solve_curve_expression(
                            liq_hval_dict, fragility_set_liq.curve_parameters, **pipeline_args)
                    # Convert PGD repairs to SI units
                    pgd_repairs = PipelineUtil.convert_result_unit(liq_fragility_curve.return_type["unit"], pgd_repairs)
                    num_pgd_repairs = pgd_repairs * length

                    # record results
                    if 'pipetype' in pipeline['properties']:
                        damage_result['pipeclass'] = pipeline['properties']['pipetype']
                    elif 'pipelinesc' in pipeline['properties']:
                        damage_result['pipeclass'] = pipeline['properties']['pipelinesc']
                    else:
                        damage_result['pipeclass'] = ""

                break_rate = 0.2 * pgv_repairs + 0.8 * pgd_repairs
                leak_rate = 0.8 * pgv_repairs + 0.2 * pgd_repairs
                total_repair_rate = pgd_repairs + pgv_repairs
                failure_probability = 1 - math.exp(-1.0 * break_rate * length)
                num_repairs = num_pgd_repairs + num_pgv_repairs

            ds_result['pgvrepairs'] = pgv_repairs
            ds_result['pgdrepairs'] = pgd_repairs
            ds_result['repairspkm'] = total_repair_rate
            ds_result['breakrate'] = break_rate
            ds_result['leakrate'] = leak_rate
            ds_result['failprob'] = failure_probability
            ds_result['numpgvrpr'] = num_pgv_repairs
            ds_result['numpgdrpr'] = num_pgd_repairs
            ds_result['numrepairs'] = num_repairs
            ds_result['haz_expose'] = AnalysisUtil.get_exposure_from_hazard_values(hazard_vals, hazard_type)

            damage_result['fragility_id'] = fragility_set.id
            damage_result['demandtypes'] = demand_types
            damage_result['demandunits'] = demand_units
            damage_result['hazardtype'] = hazard_type
            damage_result['hazardval'] = hazard_vals

            # Check if liquefaction is applicable
            if use_liquefaction is True \
                    and fragility_sets_liq is not None \
                    and geology_dataset_id is not None:
                damage_result['liq_fragility_id'] = fragility_sets_liq[pipeline["id"]].id
                damage_result['liqdemandtypes'] = liq_demand_types
                damage_result['liqdemandunits'] = liq_demand_units
                damage_result['liqhazval'] = liq_hazard_vals
                damage_result['liqprobability'] = liquefaction_prob
            else:
                damage_result['liq_fragility_id'] = None
                damage_result['liqdemandtypes'] = None
                damage_result['liqdemandunits'] = None
                damage_result['liqhazval'] = None
                damage_result['liqprobability'] = None

            ds_results.append(ds_result)
            damage_results.append(damage_result)

        # pipelines do not have matched mappings
        for pipeline in unmapped_pipelines:
            ds_result = dict()
            ds_result['guid'] = pipeline['properties']['guid']

            damage_result = dict()
            damage_result['guid'] = pipeline['properties']['guid']
            if 'pipetype' in pipeline['properties']:
                damage_result['pipeclass'] = pipeline['properties']['pipetype']
            elif 'pipelinesc' in pipeline['properties']:
                damage_result['pipeclass'] = pipeline['properties']['pipelinesc']
            else:
                damage_result['pipeclass'] = ""

            damage_result['fragility_id'] = None
            damage_result['demandtypes'] = None
            damage_result['demandunits'] = None
            damage_result['hazardtype'] = None
            damage_result['hazardval'] = None
            damage_result['liq_fragility_id'] = None
            damage_result['liqdemandtypes'] = None
            damage_result['liqdemandunits'] = None
            damage_result['liqhazval'] = None
            damage_result['liqhazval'] = None

            ds_results.append(ds_result)
            damage_results.append(damage_result)

        return ds_results, damage_results

    def get_spec(self):
        """Get specifications of the pipeline damage analysis.

        Returns:
            obj: A JSON object of specifications of the pipeline damage analysis.

        """
        return {
            'name': 'pipeline-damage',
            'description': 'buried pipeline damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'hazard_type',
                    'required': False,
                    'description': 'Hazard Type (e.g. earthquake)',
                    'type': str
                },
                {
                    'id': 'hazard_id',
                    'required': False,
                    'description': 'Hazard ID',
                    'type': str
                },
                {
                    'id': 'fragility_key',
                    'required': False,
                    'description': 'Fragility key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'use_liquefaction',
                    'required': False,
                    'description': 'Use liquefaction',
                    'type': bool
                },
                {
                    'id': 'liquefaction_fragility_key',
                    'required': False,
                    'description': 'Fragility key to use in liquefaction mapping dataset',
                    'type': str
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                },
                {
                    'id': 'liquefaction_geology_dataset_id',
                    'required': False,
                    'description': 'Geology dataset id',
                    'type': str,
                }
            ],
            'input_hazards': [
                {
                    'id': 'hazard',
                    'required': False,
                    'description': 'Hazard object',
                    'type': ["earthquake", "tsunami"]
                },
            ],
            'input_datasets': [
                {
                    'id': 'pipeline',
                    'required': True,
                    'description': 'Pipeline Inventory',
                    'type': ['ergo:buriedPipelineTopology', 'ergo:pipeline'],
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'pipeline',
                    'type': 'ergo:pipelineDamageVer3'
                },
                {
                    'id': 'metadata',
                    'parent_type': 'pipeline',
                    'description': 'additional metadata in json file about applied hazard value and '
                                   'fragility',
                    'type': 'incore:pipelineDamageSupplement'
                }
            ]
        }



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import copy
import networkx as nx


def ipw_search(v, e):
    """
    Indpendent pathway search
    :param v: vertex
    :param e: edge
    :return: path and path length
    """

    g_local = nx.Graph()

    nodelist = copy.deepcopy(v)
    edgeslist = copy.deepcopy(list(e))

    g_local.add_nodes_from(nodelist)
    for headnode, tailnode in edgeslist:
        g_local.add_edge(headnode, tailnode)

    # calculate length of each link which is 1 for all edges
    length = {}

    for (i, j) in edgeslist:
        length[i, j] = 1
        length[j, i] = 1

    # input the intact links information
    nodespair = []
    for i in nodelist:
        for j in nodelist:
            if i != j and (j, i) not in nodespair:
                nodespair.append((i, j))

    degree = {}
    degree_view = g_local.degree(nodelist)
    for pair in degree_view:
        i = pair[0]
        j = pair[1]
        degree[i] = j

    # calculate upper bound
    ub = {}

    # k-th independent path between nodes pair
    ipath = {}

    # the length of kth independent path between node pair
    path_length = {}

    for (w, q) in nodespair:

        # creat a temp list to search path
        temp_edgelist = copy.deepcopy(edgeslist)

        # up bound of possible number of independent paths
        ub[w, q] = min(degree[w], degree[q])
        ipath[w, q] = {}
        path_length[w, q] = {}
        k = 1

        # label whether stop the calculation
        flag = 1
        while flag == 1:
            g_local.clear()
            g_local.add_nodes_from(nodelist)

            for headnode, tailnode in temp_edgelist:
                g_local.add_edge(headnode, tailnode, length=length[headnode,
                                                                   tailnode])

            try:
                temp = copy.deepcopy(nx.shortest_path(g_local,
                                                      source=w,
                                                      target=q,
                                                      weight='length'))
            except nx.NetworkXNoPath as e:
                # print(w,q)
                # print("NetworkXNoPath")
                temp = []

            # if there is a path connecting the source and target,
            # start to calculate IPW
            if temp:

                # find the shortest path
                ipath[w, q][k] = copy.deepcopy(temp)
                path_length[w, q][k] = 0
                ipathtuple = []

                if len(ipath[w, q][k]) == 2:
                    # for the path just has two nodes
                    # (origin and destination)
                    ipathtuple.append((ipath[w, q][k][0],
                                       ipath[w, q][k][1]))
                    path_length[w, q][k] = length[ipath[w, q][k][0],
                                                  ipath[w, q][k][1]]

                else:
                    # for the path has more than two nodes
                    for p in range(0, len(ipath[w, q][k]) - 1):
                        ipathtuple.append((ipath[w, q][k][p],
                                           ipath[w, q][k][p + 1]))

                        path_length[w, q][k] += length[ipath[w, q][k][p],
                                                       ipath[w, q][k][p + 1]]

                # delete edges that used in previous shortest paths
                for (s, t) in ipathtuple:
                    if (s, t) in temp_edgelist:
                        temp_edgelist.remove((s, t))
                        # temp_edgelist.remove((t, s))

                k += 1

            else:
                flag = 0

    return ipath, path_length


def tipw_index(g, l, path_adt):
    """
    caculate the TIPW index of the network
    :param g: graph
    :param l: Indpendent pathway
    :param path_adt: Adt of the path
    :return: TIPW index of the network
    """

    gnodes = g.nodes()

    # compute the normalized ADT
    normal_path_adt = {}
    for key in path_adt.keys():
        normal_path_adt[key] = {}
        for i, j in path_adt[key].items():
            normal_path_adt[key][i] = len(path_adt[key].values()) * j \
                                      / sum(path_adt[key].values())

    # compute the TIPW of node
    node_tipw = {}
    for node in gnodes:
        node_tipw[node] = 0
        for pairnode in gnodes:
            if pairnode != node:
                if (node, pairnode) in l.keys():
                    for key, value in l[node, pairnode].items():
                        node_tipw[node] \
                            += normal_path_adt[node, pairnode][key] \
                               * path_service_level_edges(g, value)
                elif (pairnode, node) in l.keys():
                    for key, value in l[pairnode, node].items():
                        node_tipw[node] \
                            += normal_path_adt[pairnode, node][key] \
                               * path_service_level_edges(g, value)

    # caculate the TIPW index
    tipw_index_val = 0
    for node in gnodes:
        # network IPW
        tipw_index_val \
            += (1 / float(len(gnodes)) * node_tipw[node]) / (len(gnodes) - 1)

    return tipw_index_val


def path_service_level_edges(g, path):
    """
    compute the service level of a path
    :param g: graph
    :param path: path
    :return: service_level
    """

    service_level = 1
    for i in range(len(path) - 1):
        service_level \
            *= (1 - g.edges[path[i], path[i + 1]]['Damage_Status'] / 4.0)
    return service_level


def path_adt_from_edges(g, path):
    """
    compute the reliabity of path from a set of edges
    :param g: graph
    :param path: path
    :return: reliability
    """

    adt = max(nx.get_edge_attributes(g, 'adt').values())
    for i in range(len(path) - 1):
        adt = min(adt, g.edges[path[i], path[i + 1]]['adt'])

    return adt



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


from pyincore.analyses.meandamage.meandamage import MeanDamage



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


class BridgeUtil:
    """Utility methods for the bridge damage analysis."""
    BRIDGE_FRAGILITY_KEYS = {
        "elastomeric bearing retrofit fragility id code": [
            "Elastomeric Bearing", "eb"],
        "steel jacket retrofit fragility id code": ["Steel Jacket", "sj"],
        "restrainer cables retrofit fragility id code": ["Restrainer Cables",
                                                         "rc"],
        "seat extender retrofit fragility id code": ["Seat Extender", "se"],
        "shear key retrofit fragility id code": ["Shear Key", "sk"],
        "non-retrofit fragility id code": ["as built", "none"],
        "non-retrofit inundationdepth fragility id code": ["as built", "none"]
    }

    DEFAULT_FRAGILITY_KEY = "Non-Retrofit Fragility ID Code"
    DEFAULT_TSUNAMI_HMAX_FRAGILITY_KEY = "Non-Retrofit inundationDepth Fragility ID Code"
    GROUND_FAILURE_PROB = "groundFailureProb"

    @staticmethod
    def get_retrofit_cost(target_fragility_key):
        """Calculates retrofit cost estimate of a bridge.

        Args:
            target_fragility_key (str): Fragility key describing the type of fragility.

        Note:
            This function is not completed yet. Need real data example on the following variable
            private FeatureDataset bridgeRetrofitCostEstimate

        Returns:
            float: Retrofit cost estimate.

        """
        retrofit_cost = 0.0
        if target_fragility_key.lower() == BridgeUtil.DEFAULT_FRAGILITY_KEY.lower():
            return retrofit_cost
        else:
            retrofit_code = BridgeUtil.get_retrofit_code(target_fragility_key)
        return retrofit_cost

    @staticmethod
    def get_retrofit_type(target_fragility_key):
        """Get retrofit type by looking up BRIDGE_FRAGILITY_KEYS dictionary.

        Args:
            target_fragility_key (str): Fragility key describing the type of fragility.

        Returns:
            str: A retrofit type.

        """
        return BridgeUtil.BRIDGE_FRAGILITY_KEYS[target_fragility_key.lower()][
            0] \
            if target_fragility_key.lower() in BridgeUtil.BRIDGE_FRAGILITY_KEYS else "none"

    @staticmethod
    def get_retrofit_code(target_fragility_key):
        """Get retrofit code by looking up BRIDGE_FRAGILITY_KEYS dictionary.

        Args:
            target_fragility_key (str): Fragility key describing the type of fragility.

        Returns:
            str: A retrofit code.

        """
        return BridgeUtil.BRIDGE_FRAGILITY_KEYS[target_fragility_key.lower()][
            1] \
            if target_fragility_key.lower() in BridgeUtil.BRIDGE_FRAGILITY_KEYS else "none"
